#!/usr/bin/env python
#
# Copyright (c) 2017 10X Genomics, Inc. All rights reserved.
#
from __future__ import absolute_import

import bmkcc.cellranger.analysis.constants as analysis_constants
import bmkcc.cellranger.webshim.constants.shared as shared
import bmkcc.cellranger.rna.library as rna_library

REPORT_PREFIX_CRISPR = rna_library.get_library_type_metric_prefix(rna_library.CRISPR_LIBRARY_TYPE)
DISPLAY_PREFIX_CRISPR = rna_library.get_library_type_report_prefix(rna_library.CRISPR_LIBRARY_TYPE)

REPORT_PREFIX_ANTIBODY = rna_library.get_library_type_metric_prefix(
    rna_library.ANTIBODY_LIBRARY_TYPE
)
DISPLAY_PREFIX_ANTIBODY = rna_library.get_library_type_report_prefix(
    rna_library.ANTIBODY_LIBRARY_TYPE
)

REPORT_PREFIX_MULTIPLEXING = rna_library.get_library_type_metric_prefix(
    rna_library.MULTIPLEXING_LIBRARY_TYPE
)
DISPLAY_PREFIX_MULTIPLEXING = rna_library.get_library_type_report_prefix(
    rna_library.MULTIPLEXING_LIBRARY_TYPE
)

REPORT_PREFIX_CUSTOM = rna_library.get_library_type_metric_prefix(rna_library.CUSTOM_LIBRARY_TYPE)
DISPLAY_PREFIX_CUSTOM = rna_library.get_library_type_report_prefix(rna_library.CUSTOM_LIBRARY_TYPE)

MULTIPLEXING_NAME = "Tag"

PCA_PRCT_CLIP = [0.1, 99.9]
TSNE_TOTALCOUNTS_PRCT_CLIP = [5, 95]
TSNE_TOTALCOUNTS_DESCRIPTION = "UMI counts"
TSNE_GEMGROUPS_DESCRIPTION = "Gem group"
TSNE_CLUSTER_DESCRIPTION = ""
MAX_TOP_N_GENES = 50
MAX_DE_TABLE_ENTRIES = 10000
TOP_DE_GENES_MIN_MEAN = 1.0
MAX_WEBSHIM_BCS_DIM = 10000
MAX_WEBSHIM_KMEANS_K = 20
CLUSTERS_FILTER_TITLE = "Clustering Type"
PVALUE_DEEMPHASIS_CUTOFF = 0.1
SATURATION_LINE = 0.9

GEM_CALL_LABELS = [
    {"key": analysis_constants.GEM_CLASS_GENOME0, "label": "genome0", "color": "rgb(88,165,50)",},
    {"key": analysis_constants.GEM_CLASS_GENOME1, "label": "genome1", "color": "rgb(0,161,223)",},
    {
        "key": analysis_constants.GEM_CLASS_MULTIPLET,
        "label": "Multiplet",
        "color": "rgb(212,212,212)",
    },
]

TOTAL_READS_METRIC = {
    "name": "total_read_pairs",
    "display_name": "Number of Reads",
    "description": "Total number of read pairs that were assigned to this library in demultiplexing.",
    "format": "integer",
}

PRENORM_READS_METRIC = {
    "name": "pre_normalization_total_reads",
    "display_name": "Pre-Normalization Number of Reads",
    "description": "Total number of read pairs that were assigned to these libraries in demultiplexing.",
    "format": "integer",
}
POSTNORM_READS_METRIC = {
    "name": "post_normalization_total_reads",
    "display_name": "Post-Normalization Number of Reads",
    "description": "Number of read pairs after normalizing for depth among multiple libraries.",
    "format": "integer",
}

GOOD_BCS_METRIC = {
    "name": "good_bc_frac",
    "display_name": "Valid Barcodes",
    "description": "Fraction of reads with barcodes that match the whitelist after barcode correction.",
    "format": "percent",
}

GOOD_UMIS_METRIC = {
    "name": "good_umi_frac",
    "display_name": "Valid UMIs",
    "description": "Fraction of reads with valid UMIs; i.e., UMI sequences that do not contain Ns and that are not homopolymers.",
    "format": "percent",
}

GENOME_MAPPED_READS_METRIC = {
    "name": "genome_mapped_reads_frac",
    "display_name": "Reads Mapped to Genome",
    "description": "Fraction of reads that mapped to the %s genome.",
    "format": "percent",
    "prefix": "genomes",
    "hidden": "len(genomes) <= 1",
}

MULTI_GENOME_MAPPED_READS_METRIC = {
    "name": "multi_genome_mapped_reads_frac",
    "display_name": "Reads Mapped to Genome",
    "description": "Fraction of reads that mapped to the genome.",
    "format": "percent",
}

GENOME_CONF_MAPPED_READS_METRIC = {
    "name": "genome_conf_mapped_reads_frac",
    "display_name": "Reads Mapped Confidently to Genome",
    "description": "Fraction of reads that mapped uniquely to the %s genome. If a gene mapped to exonic loci from a single gene and also to non-exonic loci, it is considered uniquely mapped to one of the exonic loci.",
    "format": "percent",
    "prefix": "genomes",
    "hidden": "len(genomes) <= 1",
}

MULTI_GENOME_CONF_MAPPED_READS_METRIC = {
    "name": "multi_genome_conf_mapped_reads_frac",
    "display_name": "Reads Mapped Confidently to Genome",
    "description": "Fraction of reads that mapped uniquely to the %s genome. If a gene mapped to exonic loci from a single gene and also to non-exonic loci, it is considered uniquely mapped to one of the exonic loci.",
    "format": "percent",
}

TRANSCRIPTOME_CONF_MAPPED_READS_METRIC = {
    "name": "transcriptome_conf_mapped_reads_frac",
    "display_name": "Reads Mapped Confidently to Transcriptome",
    "description": "Fraction of reads that mapped to a unique gene in the %s transcriptome. The read must be consistent with annotated splice junctions. These reads are considered for UMI counting.",
    "format": "percent",
    "prefix": "genomes",
    "hidden": "len(genomes) <= 1",
}

MULTI_TRANSCRIPTOME_CONF_MAPPED_READS_METRIC = {
    "name": "multi_transcriptome_conf_mapped_reads_frac",
    "display_name": "Reads Mapped Confidently to Transcriptome",
    "description": "Fraction of reads that mapped to a unique gene in the transcriptome. The read must be consistent with annotated splice junctions. These reads are considered for UMI counting.",
    "format": "percent",
}

TARGETED_TRANSCRIPTOME_CONF_MAPPED_READS_METRIC = {
    "name": "multi_frac_conf_transcriptomic_reads_on_target",
    "display_name": "Reads Mapped Confidently to the Targeted Transcriptome",
    "description": "Fraction of reads that mapped to a unique and targeted gene in the transcriptome. The read must be consistent with annotated splice junctions. These reads are considered for UMI counting.",
    "format": "percent",
}

OFF_TARGET_TRANSCRIPTOME_CONF_MAPPED_READS_METRIC = {
    "name": "multi_frac_conf_transcriptomic_reads_off_target",
    "display_name": "Reads Mapped Confidently to the Non-Targeted Transcriptome",
    "description": "Fraction of reads that mapped to a unique and non-targeted gene in the transcriptome. The read must be consistent with annotated splice junctions. These reads are considered for UMI counting.",
    "format": "percent",
}

ANTISENSE_CONF_MAPPED_READS_METRIC = {
    "name": "multi_antisense_reads_frac",
    "display_name": "Reads Mapped Antisense to Gene",
    "description": "Fraction of reads confidently mapped to the transcriptome, but on the opposite strand of their annotated gene. A read is counted as antisense if it has any alignments that are consistent with an exon of a transcript but antisense to it, and has no sense alignments.",
    "format": "percent",
}

INTERGENIC_CONF_MAPPED_READS_METRIC = {
    "name": "intergenic_conf_mapped_reads_frac",
    "display_name": "Reads Mapped Confidently to Intergenic Regions",
    "description": "Fraction of reads that mapped uniquely to an intergenic region of the %s genome.",
    "format": "percent",
    "prefix": "genomes",
    "hidden": "len(genomes) <= 1",
}

MULTI_INTERGENIC_CONF_MAPPED_READS_METRIC = {
    "name": "multi_intergenic_conf_mapped_reads_frac",
    "display_name": "Reads Mapped Confidently to Intergenic Regions",
    "description": "Fraction of reads that mapped uniquely to an intergenic region of the genome.",
    "format": "percent",
}

INTRONIC_CONF_MAPPED_READS_METRIC = {
    "name": "intronic_conf_mapped_reads_frac",
    "display_name": "Reads Mapped Confidently to Intronic Regions",
    "description": "Fraction of reads that mapped uniquely to an intronic region of the %s genome.",
    "format": "percent",
    "prefix": "genomes",
    "hidden": "len(genomes) <= 1",
}

MULTI_INTRONIC_CONF_MAPPED_READS_METRIC = {
    "name": "multi_intronic_conf_mapped_reads_frac",
    "display_name": "Reads Mapped Confidently to Intronic Regions",
    "description": "Fraction of reads that mapped uniquely to an intronic region of the genome.",
    "format": "percent",
}

EXONIC_CONF_MAPPED_READS_METRIC = {
    "name": "exonic_conf_mapped_reads_frac",
    "display_name": "Reads Mapped Confidently to Exonic Regions",
    "description": "Fraction of reads that mapped uniquely to an exonic region of the %s genome.",
    "format": "percent",
    "prefix": "genomes",
    "hidden": "len(genomes) <= 1",
}

MULTI_EXONIC_CONF_MAPPED_READS_METRIC = {
    "name": "multi_exonic_conf_mapped_reads_frac",
    "display_name": "Reads Mapped Confidently to Exonic Regions",
    "description": "Fraction of reads that mapped uniquely to an exonic region of the genome.",
    "format": "percent",
}

NUMBER_OF_DETECTED_CELLS_METRIC = {
    "name": "filtered_bcs",
    "display_name": "Estimated Number of Cell Partitions",
    "description": "The number of barcodes associated with %s cell-containing partitions, estimated from the barcode UMI count distribution.",
    "format": "integer",
    "prefix": "genomes",
    "hidden": "len(genomes) <= 1",
}

MULTI_NUMBER_OF_DETECTED_CELLS_METRIC = {
    "name": "filtered_bcs_transcriptome_union",
    "display_name": "Estimated Number of Cells",
    "description": "The total number of barcodes associated with %s cell-containing partitions, estimated from the barcode count distribution.",
    "format": "integer",
}

MULTI_NUMBER_OF_ANTIBODY_DETECTED_CELLS_METRIC = {
    "name": REPORT_PREFIX_ANTIBODY + "filtered_bcs_transcriptome_union",
    "display_name": "Estimated Number of Cells",
    "description": "The total number of barcodes associated with %s cell-containing partitions, estimated from the barcode antibody count distribution.",
    "format": "integer",
}

READS_PER_DETECTED_CELL_METRIC = {
    "name": "multi_transcriptome_total_raw_reads_per_filtered_bc",
    "display_name": "Mean Reads per Cell",
    "description": "The total number of sequenced reads divided by the number of barcodes associated with cell-containing partitions.",
    "format": "integer",
}

TARGETED_READS_PER_DETECTED_CELL_METRIC = {
    "name": "total_targeted_reads_per_filtered_bc",
    "display_name": "Mean Targeted Reads per Cell",
    "description": "The total number of targeted reads in cells divided by the number of barcodes associated with cell-containing partitions.",
    "format": "integer",
}

PRENORM_READS_PER_CELL_METRIC = {
    "name": "pre_normalization_multi_transcriptome_total_raw_reads_per_filtered_bc",
    "display_name": "Pre-Normalization Mean Reads per Cell",
    "description": "The total number of sequenced reads divided by the number of barcodes associated with cell-containing partitions.",
    "format": "integer",
}
POSTNORM_READS_PER_CELL_METRIC = {
    "name": "post_normalization_multi_transcriptome_total_raw_reads_per_filtered_bc",
    "display_name": "Post-Normalization Mean Reads per Cell",
    "description": "The total number of (depth-normalized) reads divided by the number of barcodes associated with cell-containing partitions.",
    "format": "integer",
}

GENES_PER_DETECTED_CELL_METRIC = {
    "name": "filtered_bcs_median_unique_genes_detected",
    "display_name": "Median Genes per Cell",
    "description": "The median number of genes detected per %s cell-associated barcode. Detection is defined as the presence of at least 1 UMI count.",
    "format": "integer",
    "prefix": "genomes",
    "hidden": "target_set is not None",
}

TARGETED_GENES_PER_DETECTED_CELL_METRIC = {
    "name": "median_genes_per_cell_on_target",
    "display_name": "Median Targeted Genes per Cell",
    "description": "The median number of targeted genes detected per cell-associated barcode. Detection is defined as the presence of at least 1 UMI count.",
    "format": "integer",
}

COUNTS_PER_DETECTED_CELL_METRIC = {
    "name": "filtered_bcs_median_counts",
    "display_name": "Median UMI Counts per Cell",
    "description": "The median number of UMI counts per %s cell-associated barcode.",
    "format": "integer",
    "prefix": "genomes",
    "hidden": "target_set is not None",
}

TARGETED_COUNTS_PER_DETECTED_CELL_METRIC = {
    "name": "median_umis_per_cell_on_target",
    "display_name": "Median Targeted UMI Counts per Cell",
    "description": "The median number of Targeted UMI counts per cell-associated barcode.",
    "format": "integer",
}

# Deprecated
CDNA_DUPE_FRAC_METRIC = {
    "name": "multi_cdna_pcr_dupe_reads_frac",
    "display_name": "cDNA PCR Duplication",
    "description": "The fraction of candidate reads that can be attributed to the cDNA amplification process; this is the number of non-unique candidate reads divided by the total number of candidate reads. A read is a candidate if it is confidently mapped and has a valid UMI and cell barcode.",
    "format": "percent",
    "hidden": "True",
}
SEQUENCING_SATURATION_METRIC = {
    "name": "multi_cdna_pcr_dupe_reads_frac",
    "display_name": "Sequencing Saturation",
    "description": 'The fraction of reads originating from an already-observed UMI. This is a function of library complexity and sequencing depth. More specifically, this is the fraction of confidently mapped, valid cell-barcode, valid UMI reads that had a non-unique (cell-barcode, UMI, gene). This metric was called "cDNA PCR Duplication" in versions of Cell Ranger prior to 1.2.',
    "format": "percent",
    "hidden": "target_set is not None",
}
TARGETED_SEQUENCING_SATURATION_METRIC = {
    "name": "multi_cdna_pcr_dupe_reads_frac_on_target",
    "display_name": "Targeted Sequencing Saturation",
    "description": 'The fraction of targeted reads originating from an already-observed targeted UMI. This is a function of library complexity and sequencing depth. More specifically, this is the fraction of confidently mapped, valid cell-barcode, valid targeted UMI reads that had a non-unique (cell-barcode, UMI, gene). This metric was called "cDNA PCR Duplication" in versions of Cell Ranger prior to 1.2.',
    "format": "percent",
    "hidden": "target_set is None",
}

SUMMARY_GENES_PER_DETECTED_CELL_METRIC = GENES_PER_DETECTED_CELL_METRIC.copy()
SUMMARY_GENES_PER_DETECTED_CELL_METRIC["hidden"] = "len(genomes) > 1 or target_set is not None"

# These ultimately get displayed
SUMMARY_METRICS = [
    MULTI_NUMBER_OF_DETECTED_CELLS_METRIC,
    READS_PER_DETECTED_CELL_METRIC,
    POSTNORM_READS_PER_CELL_METRIC,
    SUMMARY_GENES_PER_DETECTED_CELL_METRIC,
]

MULTIPLET_RATE_METRIC = {
    "name": "filtered_bcs_inferred_multiplet_rate",
    "display_name": "Fraction GEMs with >1 Cell",
    "description": "The mean fraction of cell-associated barcodes estimated to be associated with more than one cell, calculated via bootstrap sampling and adjusting for the ratio of the two cell types.",
    "format": "percent",
}

MULTIPLET_RATE_LB_METRIC = {
    "name": "filtered_bcs_inferred_multiplet_rate_lb",
    "display_name": "Fraction GEMs with >1 Cell (Lower Bound)",
    "description": "The lower bound of the 95% confidence interval of the fraction of cell-associated barcodes estimated to be associated with more than one cell, calculated via bootstrap sampling and adjusting for the ratio of the two cell types.",
    "format": "percent",
}

MULTIPLET_RATE_UB_METRIC = {
    "name": "filtered_bcs_inferred_multiplet_rate_ub",
    "display_name": "Fraction GEMs with >1 Cell (Upper Bound)",
    "description": "The upper bound of the 95% confidence interval of the fraction of cell-associated barcodes estimated to be associated with more than one cell, calculated via bootstrap sampling and adjusting for the ratio of the two cell types.",
    "format": "percent",
}

MEAN_COUNT_PURITY_METRIC = {
    "name": "multi_filtered_bcs_mean_count_purity",
    "display_name": "Mean UMI Count Purity",
    "description": "Among single-cell GEM barcodes, the total fraction of UMI counts coming from the transcriptome of the cell inferred to be in the GEM, as opposed to a transcriptome that should not be present in the cell.",
    "format": "percent",
}

FRAC_READS_IN_CELLS_METRIC = {
    "name": "filtered_bcs_conf_mapped_barcoded_reads_cum_frac",
    "display_name": "Fraction Reads in Cells",
    "description": "The fraction of valid-barcode, confidently-mapped-to-transcriptome reads with %s cell-associated barcodes.",
    "format": "percent",
    "prefix": "genomes",
    "hidden": "len(genomes) <= 1",
}

# Same as above, but summarized across all genomes present
MULTI_FRAC_READS_IN_CELLS_METRIC = {
    "name": "multi_filtered_bcs_conf_mapped_barcoded_reads_cum_frac",
    "display_name": "Fraction Reads in Cells",
    "description": "The fraction of valid-barcode, confidently-mapped-to-transcriptome reads with cell-associated barcodes.",
    "format": "percent",
}

MEAN_USABLE_READS_IN_CELLS_METRIC = {
    "name": "filtered_bcs_conf_mapped_barcoded_reads_per_filtered_bc",
    "display_name": "Mean Usable Reads per Cell",
    "description": "The total number of reads that have a valid barcode, valid UMI, confidently map to the transcriptome and are associated with a cell-containing GEM, divided by the number of cells.",
    "format": "integer",
    "prefix": "genomes",
}


TOTAL_GENES_DETECTED_METRIC = {
    "name": "filtered_bcs_total_unique_genes_detected",
    "display_name": "Total Genes Detected",
    "description": "The number of %s genes with at least one UMI count in any cell.",
    "format": "integer",
    "prefix": "genomes",
    "hidden": "target_set is not None",
}

TOTAL_TARGETED_GENES_DETECTED_METRIC = {
    "name": "num_genes_detected_on_target",
    "display_name": "Total Targeted Genes Detected",
    "description": "The number of targeted genes with at least one UMI count in any cell.",
    "format": "integer",
}

Q30_METRICS = [
    {
        "name": "bc_bases_with_q30_frac",
        "display_name": "Q30 Bases in Barcode",
        "description": "Fraction of cell barcode bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
    {
        "name": "read_bases_with_q30_frac",
        "display_name": "Q30 Bases in RNA Read",
        "description": "Fraction of RNA read bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator. This is Read 1 for the Single Cell 3' v1 chemistry and Read 2 for the Single Cell 3' v2 chemistry.",
        "format": "percent",
    },
    {
        "name": "read2_bases_with_q30_frac",
        "display_name": "Q30 Bases in RNA Read 2",
        "description": "Fraction of RNA read 2 bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
    {
        "name": "umi_bases_with_q30_frac",
        "display_name": "Q30 Bases in UMI",
        "description": "Fraction of UMI bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
]

GEX_DUAL_INDEXED_Q30_METRICS = [
    {
        "name": "i1_bases_with_q30_frac",
        "display_name": "Q30 Bases in Sample Index I1",
        "description": "Fraction of I1 (or I7) read bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
    {
        "name": "i2_bases_with_q30_frac",
        "display_name": "Q30 Bases in Sample Index I2",
        "description": "Fraction of I2 (or I5) read bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
]

CRISPR_Q30_METRICS = [
    {
        "name": REPORT_PREFIX_CRISPR + "bc_bases_with_q30_frac",
        "display_name": DISPLAY_PREFIX_CRISPR + "Q30 Bases in Barcode",
        "description": "Fraction of cell barcode bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_CRISPR + "read_bases_with_q30_frac",
        "display_name": DISPLAY_PREFIX_CRISPR + "Q30 Bases in CRISPR Read",
        "description": "Fraction of CRISPR library read bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator. This is Read 2 for the Single Cell 3' v3 and Single Cell 5' chemistries.",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_CRISPR + "read2_bases_with_q30_frac",
        "display_name": DISPLAY_PREFIX_CRISPR + "Q30 Bases in CRISPR Read 2",
        "description": "Fraction of CRISPR library read 2 bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_CRISPR + "umi_bases_with_q30_frac",
        "display_name": DISPLAY_PREFIX_CRISPR + "Q30 Bases in UMI",
        "description": "Fraction of CRISPR library UMI bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
]

CRISPR_SEQUENCING_METRICS = [
    {
        "name": REPORT_PREFIX_CRISPR + "total_read_pairs",
        "display_name": DISPLAY_PREFIX_CRISPR + "Number of Reads",
        "description": "Total number of CRISPR library reads",
        "format": "integer",
    },
    {
        "name": REPORT_PREFIX_CRISPR + "reads_per_cell",
        "display_name": DISPLAY_PREFIX_CRISPR + "Mean Reads per Cell",
        "description": "The total number of sequenced CRISPR library reads divided by the number of barcodes associated with cell-containing partitions.",
        "format": "integer",
    },
    {
        "name": REPORT_PREFIX_CRISPR + "good_bc_frac",
        "display_name": DISPLAY_PREFIX_CRISPR + "Valid Barcodes",
        "description": "Fraction of CRISPR library reads with a barcode found in or corrected to one that is found in the whitelist",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_CRISPR + "multi_cdna_pcr_dupe_reads_frac",
        "display_name": DISPLAY_PREFIX_CRISPR + "Sequencing Saturation",
        "description": "The fraction of CRISPR library reads originating from an already-observed UMI. This is a function of library complexity and sequencing depth. More specifically, this is the fraction of confidently mapped, valid cell-barcode, valid UMI reads that had a non-unique (cell-barcode, UMI, CRISPR feature barcode).",
        "format": "percent",
    },
] + CRISPR_Q30_METRICS

CRISPR_APPLICATION_METRICS = [
    {
        "name": REPORT_PREFIX_CRISPR + "feature_bc_extracted_frac",
        "display_name": DISPLAY_PREFIX_CRISPR + "Fraction Reads with Putative Protospacer Sequence",
        "description": "Fraction of CRISPR library reads from which a putative protospacer sequence could be extracted",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_CRISPR + "recognized_feature_bc_frac",
        "display_name": DISPLAY_PREFIX_CRISPR + "Fraction Guide Reads",
        "description": "Fraction of CRISPR library reads with a recognized protospacer sequence",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_CRISPR + "frac_feature_reads_usable",
        "display_name": DISPLAY_PREFIX_CRISPR + "Fraction Guide Reads Usable",
        "description": "Fraction of CRISPR library reads with a recognized protospacer sequence, a valid UMI, and a cell-associated barcode",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_CRISPR + "feature_reads_usable_per_cell",
        "display_name": DISPLAY_PREFIX_CRISPR + "Guide Reads Usable per Cell",
        "description": "Number of CRISPR library guide reads usable divided by the number of cell-associated barcodes",
        "format": "integer",
    },
    {
        "name": REPORT_PREFIX_CRISPR + "unrecognized_feature_bc_frac",
        "display_name": DISPLAY_PREFIX_CRISPR + "Fraction Protospacer Not Recognized",
        "description": "Among all CRISPR library reads with a putative protospacer sequence, the fraction with a protospacer sequence that was not recognized",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_CRISPR + "feature_reads_in_cells",
        "display_name": DISPLAY_PREFIX_CRISPR + "Guide Reads in Cells",
        "description": "Among CRISPR library reads with a recognized protospacer sequence, a valid UMI, and a valid barcode, the fraction associated with cell-containing partitions",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_CRISPR + "frac_cells_with_protospacer",
        "display_name": DISPLAY_PREFIX_CRISPR + "Cells with 1 or more protospacers detected",
        "description": "Cells with 1 or more protospacers detected",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_CRISPR + "frac_cells_with_multiple_protospacer",
        "display_name": DISPLAY_PREFIX_CRISPR + "Cells with 2 or more protospacers detected",
        "description": "Cells with 2 or more protospacers detected",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_CRISPR + "multi_filtered_bcs_median_counts",
        "display_name": DISPLAY_PREFIX_CRISPR
        + "Median UMIs per Cell (summed over all recognized protospacers)",
        "description": "Median UMIs per Cell (summed over all recognized protospacers)",
        "format": "integer",
    },
]


MULTIPLEXING_APPLICATION_METRICS = [
    {
        "name": REPORT_PREFIX_MULTIPLEXING + "recognized_feature_bc_frac",
        "display_name": DISPLAY_PREFIX_MULTIPLEXING + "Fraction {} Reads".format(MULTIPLEXING_NAME),
        "description": "Fraction of {name} library reads with a recognized {name} sequence".format(
            name=MULTIPLEXING_NAME
        ),
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_MULTIPLEXING + "frac_feature_reads_usable",
        "display_name": DISPLAY_PREFIX_MULTIPLEXING
        + "Fraction {} Reads Usable".format(MULTIPLEXING_NAME),
        "description": "Fraction of {name} library reads with a recognized {name} sequence, a valid UMI, and a cell-associated barcode".format(
            name=MULTIPLEXING_NAME
        ),
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_MULTIPLEXING + "feature_reads_usable_per_cell",
        "display_name": DISPLAY_PREFIX_MULTIPLEXING
        + "{} Reads Usable per Cell".format(MULTIPLEXING_NAME),
        "description": "Number of {} library guide reads usable divided by the number of cell-associated barcodes".format(
            MULTIPLEXING_NAME
        ),
        "format": "integer",
    },
    {
        "name": REPORT_PREFIX_MULTIPLEXING + "unrecognized_feature_bc_frac",
        "display_name": DISPLAY_PREFIX_MULTIPLEXING
        + "Fraction {} Not Recognized".format(MULTIPLEXING_NAME),
        "description": "Among all {name} library reads, the fraction with a {name} sequence that was not recognized".format(
            name=MULTIPLEXING_NAME
        ),
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_MULTIPLEXING + "feature_reads_in_cells",
        "display_name": DISPLAY_PREFIX_MULTIPLEXING + "{} Reads in Cells".format(MULTIPLEXING_NAME),
        "description": "Among {name} library reads with a recognized {name} sequence, a valid UMI, and a valid barcode, the fraction associated with cell-containing partitions".format(
            name=MULTIPLEXING_NAME
        ),
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_MULTIPLEXING + "frac_cells_with_{}".format(MULTIPLEXING_NAME),
        "display_name": DISPLAY_PREFIX_MULTIPLEXING
        + "Cells with 1 or more {} assigned".format(MULTIPLEXING_NAME),
        "description": "Cells with 1 or more {} assigned".format(MULTIPLEXING_NAME),
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_MULTIPLEXING
        + "frac_cells_with_multiple_{}".format(MULTIPLEXING_NAME),
        "display_name": DISPLAY_PREFIX_MULTIPLEXING
        + "Cells with 2 or more {} assigned".format(MULTIPLEXING_NAME),
        "description": "Cells with 2 or more {} assigned".format(MULTIPLEXING_NAME),
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_MULTIPLEXING + "multi_filtered_bcs_median_counts",
        "display_name": DISPLAY_PREFIX_MULTIPLEXING
        + "Median UMIs per Cell (summed over all recognized {})".format(MULTIPLEXING_NAME),
        "description": "Median UMIs per Cell (summed over all recognized {})".format(
            MULTIPLEXING_NAME
        ),
        "format": "integer",
    },
]

ANTIBODY_CELL_CALLING_METRICS = [MULTI_NUMBER_OF_ANTIBODY_DETECTED_CELLS_METRIC]

ANTIBODY_Q30_METRICS = [
    {
        "name": REPORT_PREFIX_ANTIBODY + "bc_bases_with_q30_frac",
        "display_name": DISPLAY_PREFIX_ANTIBODY + "Q30 Bases in Barcode",
        "description": "Fraction of Antibody library cell barcode bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_ANTIBODY + "read_bases_with_q30_frac",
        "display_name": DISPLAY_PREFIX_ANTIBODY + "Q30 Bases in Antibody Read",
        "description": "Fraction of Antibody library read bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator. This is Read 2 for the Single Cell 3' v3 and Single Cell 5' chemistries.",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_ANTIBODY + "read2_bases_with_q30_frac",
        "display_name": DISPLAY_PREFIX_ANTIBODY + "Q30 Bases in Antibody Read 2",
        "description": "Fraction of Antibody library read 2 bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_ANTIBODY + "umi_bases_with_q30_frac",
        "display_name": DISPLAY_PREFIX_ANTIBODY + "Q30 Bases in UMI",
        "description": "Fraction of Antibody library UMI bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
]


ANTIBODY_SEQUENCING_METRICS = [
    {
        "name": REPORT_PREFIX_ANTIBODY + "total_read_pairs",
        "display_name": DISPLAY_PREFIX_ANTIBODY + "Number of Reads",
        "description": "Total number of Antibody library reads",
        "format": "integer",
    },
    {
        "name": REPORT_PREFIX_ANTIBODY + "reads_per_cell",
        "display_name": DISPLAY_PREFIX_ANTIBODY + "Mean Reads per Cell",
        "description": "The total number of sequenced Antibody library reads divided by the number of barcodes associated with cell-containing partitions.",
        "format": "integer",
    },
    {
        "name": REPORT_PREFIX_ANTIBODY + "good_bc_frac",
        "display_name": DISPLAY_PREFIX_ANTIBODY + "Valid Barcodes",
        "description": "Fraction of Antibody library reads with a barcode found in or corrected to one that is found in the whitelist",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_ANTIBODY + "multi_cdna_pcr_dupe_reads_frac",
        "display_name": DISPLAY_PREFIX_ANTIBODY + "Sequencing Saturation",
        "description": "The fraction of Antibody library reads originating from an already-observed UMI. This is a function of library complexity and sequencing depth. More specifically, this is the fraction of confidently mapped, valid cell-barcode, valid UMI reads that had a non-unique (cell-barcode, UMI, antibody feature barcode).",
        "format": "percent",
    },
] + ANTIBODY_Q30_METRICS

ANTIBODY_APPLICATION_METRICS = [
    {
        "name": REPORT_PREFIX_ANTIBODY + "recognized_feature_bc_frac",
        "display_name": DISPLAY_PREFIX_ANTIBODY + "Fraction Antibody Reads",
        "description": "Fraction of Antibody library reads that contain a recognized antibody barcode",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_ANTIBODY + "frac_feature_reads_usable",
        "display_name": DISPLAY_PREFIX_ANTIBODY + "Fraction Antibody Reads Usable",
        "description": "Fraction of Antibody library reads that contain a recognized antibody barcode, a valid UMI, and a cell-associated barcode",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_ANTIBODY + "feature_reads_usable_per_cell",
        "display_name": DISPLAY_PREFIX_ANTIBODY + "Antibody Reads Usable per Cell",
        "description": "Number of Antibody library reads usable divided by the number of cell-associated barcodes",
        "format": "integer",
    },
    {
        "name": REPORT_PREFIX_ANTIBODY + "reads_lost_to_aggregate_GEMs",
        "display_name": DISPLAY_PREFIX_ANTIBODY + "Fraction Antibody Reads in Aggregate Barcodes",
        "description": "Fraction of Antibody library reads that were lost after removing aggregate barcodes.",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_ANTIBODY + "unrecognized_feature_bc_frac",
        "display_name": DISPLAY_PREFIX_ANTIBODY + "Fraction Unrecognized Antibody",
        "description": "Among all Antibody library reads, the fraction with an unrecognizable antibody barcode",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_ANTIBODY + "feature_reads_in_cells",
        "display_name": DISPLAY_PREFIX_ANTIBODY + "Antibody Reads in Cells",
        "description": "Among Antibody library reads with a recognized antibody barcode, a valid UMI, and a valid barcode, the fraction associated with cell-containing partitions",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_ANTIBODY + "multi_filtered_bcs_median_counts",
        "display_name": DISPLAY_PREFIX_ANTIBODY
        + "Median UMIs per Cell (summed over all recognized antibody barcodes)",
        "description": "Median UMIs per Cell (summed over all recognized antibody barcodes)",
        "format": "integer",
    },
]

MULTIPLEXING_Q30_METRICS = [
    {
        "name": REPORT_PREFIX_MULTIPLEXING + "bc_bases_with_q30_frac",
        "display_name": DISPLAY_PREFIX_MULTIPLEXING + "Q30 Bases in Barcode",
        "description": "Fraction of cell barcode bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_MULTIPLEXING + "read_bases_with_q30_frac",
        "display_name": DISPLAY_PREFIX_MULTIPLEXING + "Q30 Bases in MULTIPLEXING Read",
        "description": "Fraction of MULTIPLEXING library read bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator. This is Read 2 for the Single Cell 3' v3 and Single Cell 5' chemistries.",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_MULTIPLEXING + "read2_bases_with_q30_frac",
        "display_name": DISPLAY_PREFIX_MULTIPLEXING + "Q30 Bases in MULTIPLEXING Read 2",
        "description": "Fraction of MULTIPLEXING library read 2 bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_MULTIPLEXING + "umi_bases_with_q30_frac",
        "display_name": DISPLAY_PREFIX_MULTIPLEXING + "Q30 Bases in UMI",
        "description": "Fraction of MULTIPLEXING library UMI bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
]

MULTIPLEXING_SEQUENCING_METRICS = [
    {
        "name": REPORT_PREFIX_MULTIPLEXING + "total_read_pairs",
        "display_name": DISPLAY_PREFIX_MULTIPLEXING + "Number of Reads",
        "description": "Total number of MULTIPLEXING library reads",
        "format": "integer",
    },
    {
        "name": REPORT_PREFIX_MULTIPLEXING + "reads_per_cell",
        "display_name": DISPLAY_PREFIX_MULTIPLEXING + "Mean Reads per Cell",
        "description": "The total number of sequenced MULTIPLEXING library reads divided by the number of barcodes associated with cell-containing partitions.",
        "format": "integer",
    },
    {
        "name": REPORT_PREFIX_MULTIPLEXING + "good_bc_frac",
        "display_name": DISPLAY_PREFIX_MULTIPLEXING + "Valid Barcodes",
        "description": "Fraction of MULTIPLEXING library reads with a barcode found in or corrected to one that is found in the whitelist",
        "format": "percent",
    },
    {
        "name": REPORT_PREFIX_MULTIPLEXING + "multi_cdna_pcr_dupe_reads_frac",
        "display_name": DISPLAY_PREFIX_MULTIPLEXING + "Sequencing Saturation",
        "description": "The fraction of MULTIPLEXING library reads originating from an already-observed UMI. This is a function of library complexity and sequencing depth. More specifically, this is the fraction of valid cell-barcode, valid UMI, recongnized MULTIPLEXING feature barcode reads that had a non-unique (cell-barcode, UMI, MULTIPLEXING feature barcode).",
        "format": "percent",
    },
] + MULTIPLEXING_Q30_METRICS


CUSTOM_FEATURE_Q30_METRICS = [
    {
        "name": "bc_bases_with_q30_frac",
        "display_name": "%s: Q30 Bases in Barcode",
        "description": "Fraction of Custom library cell barcode bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
        "prefix": "custom_features",
    },
    {
        "name": "read_bases_with_q30_frac",
        "display_name": "%s: Q30 Bases in Feature Read",
        "description": "Fraction of Custom library read bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator. This is Read 2 for the Single Cell 3' v3 and Single Cell 5' chemistries.",
        "format": "percent",
        "prefix": "custom_features",
    },
    {
        "name": "read2_bases_with_q30_frac",
        "display_name": "%s: Q30 Bases in Feature Read 2",
        "description": "Fraction of Custom library read 2 bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
        "prefix": "custom_features",
    },
    {
        "name": "umi_bases_with_q30_frac",
        "display_name": "%s: Q30 Bases in UMI",
        "description": "Fraction of Custom library UMI bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
        "prefix": "custom_features",
    },
]


CUSTOM_FEATURE_SEQUENCING_METRICS = [
    {
        "name": "total_read_pairs",
        "display_name": "%s: Number of Reads",
        "description": "Total number of Custom library reads",
        "format": "integer",
        "prefix": "custom_features",
    },
    {
        "name": "reads_per_cell",
        "display_name": "%s: Mean Reads per Cell",
        "description": "The total number of sequenced Custom library reads divided by the number of barcodes associated with cell-containing partitions.",
        "format": "integer",
        "prefix": "custom_features",
    },
    {
        "name": "good_bc_frac",
        "display_name": "%s: Valid Barcodes",
        "description": "Fraction of Custom library reads with a barcode found in or corrected to one that is found in the whitelist",
        "format": "percent",
        "prefix": "custom_features",
    },
    {
        "name": "multi_cdna_pcr_dupe_reads_frac",
        "display_name": "%s: Sequencing Saturation",
        "description": "The fraction of Custom library reads originating from an already-observed UMI. This is a function of library complexity and sequencing depth. More specifically, this is the fraction of confidently mapped, valid cell-barcode, valid UMI reads that had a non-unique (cell-barcode, UMI, feature barcode).",
        "format": "percent",
        "prefix": "custom_features",
    },
] + CUSTOM_FEATURE_Q30_METRICS

CUSTOM_FEATURE_APPLICATION_METRICS = [
    {
        "name": "recognized_feature_bc_frac",
        "display_name": "%s: Fraction Feature Reads",
        "description": "Fraction of Custom library reads that contain a recognized feature barcode",
        "format": "percent",
        "prefix": "custom_features",
    },
    {
        "name": "frac_feature_reads_usable",
        "display_name": "%s: Fraction Feature Reads Usable",
        "description": "Fraction of Custom library reads that contain a recognized feature barcode, a valid UMI, and a cell-associated barcode",
        "format": "percent",
        "prefix": "custom_features",
    },
    {
        "name": "feature_reads_usable_per_cell",
        "display_name": "%s: Feature Reads Usable per Cell",
        "description": "Number of feature Custom library reads usable divided by the number of cell-associated barcodes",
        "format": "integer",
        "prefix": "custom_features",
    },
    {
        "name": "unrecognized_feature_bc_frac",
        "display_name": "%s: Fraction Unrecognized Feature",
        "description": "Among all Custom library reads, the fraction with an unrecognizable feature barcode",
        "format": "percent",
        "prefix": "custom_features",
    },
    {
        "name": "feature_reads_in_cells",
        "display_name": "%s: Feature Reads in Cells",
        "description": "Among Custom library reads with a recognized feature barcode, a valid UMI, and a valid barcode, the fraction associated with cell-containing partitions",
        "format": "percent",
        "prefix": "custom_features",
    },
    {
        "name": "multi_filtered_bcs_median_counts",
        "display_name": "%s: Median UMIs per Cell (summed over all recognized feature barcodes)",
        "description": "Median UMIs per Cell (summed over all recognized feature barcodes)",
        "format": "integer",
        "prefix": "custom_features",
    },
]

SEQUENCING_METRICS = [
    TOTAL_READS_METRIC,
    PRENORM_READS_METRIC,
    POSTNORM_READS_METRIC,
    GOOD_BCS_METRIC,
    CDNA_DUPE_FRAC_METRIC,
    SEQUENCING_SATURATION_METRIC,
    TARGETED_SEQUENCING_SATURATION_METRIC,
] + Q30_METRICS

MAPPING_METRICS = [
    MULTI_GENOME_MAPPED_READS_METRIC,
    GENOME_MAPPED_READS_METRIC,
    MULTI_GENOME_CONF_MAPPED_READS_METRIC,
    GENOME_CONF_MAPPED_READS_METRIC,
    MULTI_INTERGENIC_CONF_MAPPED_READS_METRIC,
    INTERGENIC_CONF_MAPPED_READS_METRIC,
    MULTI_INTRONIC_CONF_MAPPED_READS_METRIC,
    INTRONIC_CONF_MAPPED_READS_METRIC,
    MULTI_EXONIC_CONF_MAPPED_READS_METRIC,
    EXONIC_CONF_MAPPED_READS_METRIC,
    MULTI_TRANSCRIPTOME_CONF_MAPPED_READS_METRIC,
    TRANSCRIPTOME_CONF_MAPPED_READS_METRIC,
    TARGETED_TRANSCRIPTOME_CONF_MAPPED_READS_METRIC,
    OFF_TARGET_TRANSCRIPTOME_CONF_MAPPED_READS_METRIC,
    ANTISENSE_CONF_MAPPED_READS_METRIC,
]

DETECTED_CELL_METRICS = [
    MULTI_NUMBER_OF_DETECTED_CELLS_METRIC,
    NUMBER_OF_DETECTED_CELLS_METRIC,
    MULTI_FRAC_READS_IN_CELLS_METRIC,
    FRAC_READS_IN_CELLS_METRIC,
    READS_PER_DETECTED_CELL_METRIC,
    TARGETED_READS_PER_DETECTED_CELL_METRIC,
    PRENORM_READS_PER_CELL_METRIC,
    POSTNORM_READS_PER_CELL_METRIC,
    GENES_PER_DETECTED_CELL_METRIC,
    TARGETED_GENES_PER_DETECTED_CELL_METRIC,
    TOTAL_GENES_DETECTED_METRIC,
    TOTAL_TARGETED_GENES_DETECTED_METRIC,
    COUNTS_PER_DETECTED_CELL_METRIC,
    TARGETED_COUNTS_PER_DETECTED_CELL_METRIC,
]

# NOTE: these metrics only apply to barnyard samples
GEM_METRICS = [
    {
        "name": "filtered_bcs_transcriptome_union",
        "display_name": "GEMs with >0 Cells",
        "description": "The number of barcodes associated with at least one cell.",
        "format": "integer",
        "hidden": "len(genomes) <= 1",
    },
    {
        "name": "filtered_bcs_inferred_multiplets",
        "display_name": "GEMs with >1 Cell",
        "description": "The inferred number of barcodes that are associated with multiple cells.",
        "format": "integer",
    },
    MULTIPLET_RATE_METRIC,
    MULTIPLET_RATE_LB_METRIC,
    MULTIPLET_RATE_UB_METRIC,
    MEAN_COUNT_PURITY_METRIC,
]

# NOTE; these metrics only apply to aggregated samples
AGGREGATION_METRICS = [
    {
        "name": "frac_reads_kept",
        "display_name": "%s Fraction of Reads Kept",
        "description": "The fraction of reads kept for this input sample, after normalizing depth across samples to reduce batch effects.",
        "format": "percent",
        "prefix": "agg_batches",
    },
    {
        "name": "pre_normalization_raw_reads_per_filtered_bc",
        "display_name": "%s Pre-Normalization Total Reads per Cell",
        "description": "The mean number of total sequencing reads per cell in this input sample, prior to depth normalization.",
        "format": "integer",
        "prefix": "agg_batches",
    },
    {
        "name": "pre_normalization_cmb_reads_per_filtered_bc",
        "display_name": "%s Pre-Normalization Confidently Mapped Barcoded Reads per Cell",
        "description": "The mean number of confidently-mapped-to-transcriptome, valid-barcode reads per cell in this input sample, prior to depth normalization.",
        "format": "integer",
        "prefix": "agg_batches",
    },
]

CRISPR_AGGREGATION_METRICS = [
    {
        "name": REPORT_PREFIX_CRISPR + "frac_reads_kept",
        "display_name": "%s Fraction of Reads Kept",
        "description": "The fraction of reads kept for this input sample, after normalizing depth across samples to reduce batch effects.",
        "format": "percent",
        "prefix": "agg_batches",
    },
    {
        "name": REPORT_PREFIX_CRISPR + "pre_normalization_raw_reads_per_filtered_bc",
        "display_name": "%s Pre-Normalization Total Reads per Cell",
        "description": "The mean number of total sequencing reads per cell in this input sample, prior to depth normalization.",
        "format": "integer",
        "prefix": "agg_batches",
    },
    {
        "name": REPORT_PREFIX_CRISPR + "pre_normalization_cmb_reads_per_filtered_bc",
        "display_name": "%s Pre-Normalization Confidently Mapped Barcoded Reads per Cell",
        "description": "The mean number of confidently-mapped-to-transcriptome, valid-barcode reads per cell in this input sample, prior to depth normalization.",
        "format": "integer",
        "prefix": "agg_batches",
    },
]

ANTIBODY_AGGREGATION_METRICS = [
    {
        "name": REPORT_PREFIX_ANTIBODY + "frac_reads_kept",
        "display_name": "%s Fraction of Reads Kept",
        "description": "The fraction of reads kept for this input sample, after normalizing depth across samples to reduce batch effects.",
        "format": "percent",
        "prefix": "agg_batches",
    },
    {
        "name": REPORT_PREFIX_ANTIBODY + "pre_normalization_raw_reads_per_filtered_bc",
        "display_name": "%s Pre-Normalization Total Reads per Cell",
        "description": "The mean number of total sequencing reads per cell in this input sample, prior to depth normalization.",
        "format": "integer",
        "prefix": "agg_batches",
    },
    {
        "name": REPORT_PREFIX_ANTIBODY + "pre_normalization_cmb_reads_per_filtered_bc",
        "display_name": "%s Pre-Normalization Confidently Mapped Barcoded Reads per Cell",
        "description": "The mean number of confidently-mapped-to-transcriptome, valid-barcode reads per cell in this input sample, prior to depth normalization.",
        "format": "integer",
        "prefix": "agg_batches",
    },
]

CUSTOM_FEATURE_AGGREGATION_METRICS = [
    {
        "name": REPORT_PREFIX_CUSTOM + "_frac_reads_kept",
        "display_name": "%s Fraction of Reads Kept",
        "description": "The fraction of reads kept for this input sample, after normalizing depth across samples to reduce batch effects.",
        "format": "percent",
        "prefix": "agg_batches",
    },
    {
        "name": REPORT_PREFIX_CUSTOM + "_pre_normalization_raw_reads_per_filtered_bc",
        "display_name": "%s Pre-Normalization Total Reads per Cell",
        "description": "The mean number of total sequencing reads per cell in this input sample, prior to depth normalization.",
        "format": "integer",
        "prefix": "agg_batches",
    },
    {
        "name": REPORT_PREFIX_CUSTOM + "_pre_normalization_cmb_reads_per_filtered_bc",
        "display_name": "%s Pre-Normalization Confidently Mapped Barcoded Reads per Cell",
        "description": "The mean number of confidently-mapped-to-transcriptome, valid-barcode reads per cell in this input sample, prior to depth normalization.",
        "format": "integer",
        "prefix": "agg_batches",
    },
]

TARGETED_APPLICATION_METRICS = [
    {
        "name": "num_genes_on_target",
        "display_name": "Number of Targeted Genes",
        "description": "Number of Targeted Genes as specified in the input target panel file.",
        "format": "integer",
    },
    {
        "name": "num_genes_off_target",
        "display_name": "Number of Non-Targeted Genes",
        "description": "Number of genes in the reference genome that are not targeted.",
        "format": "integer",
    },
    {
        "name": "num_genes_quantifiable_on_target",
        "display_name": "Number of Targeted Genes >= 10 UMIs",
        "description": "Number of Targeted Genes with at least 10 UMIs in cell-associated barcodes. These genes are then considered in the calculation of per-gene enrichments.",
        "format": "integer",
    },
    {
        "name": "num_genes_quantifiable_off_target",
        "display_name": "Number of Non-Targeted Genes >= 10 UMIs",
        "description": "Number of Non-Targeted Genes with at least 10 UMIs in cell-associated barcodes. These genes are then considered in the calculation of per-gene enrichments.",
        "format": "integer",
    },
    {
        "name": "num_rpu_enriched_genes_on_target",
        "display_name": "Number of Enriched Targeted Genes",
        "description": "Genes are classified with a two-class Gaussian mixture model into two groups based on the mean Reads per UMI per gene. Enriched genes have higher mean Reads Per UMI, indicating they have been sequenced to much higher saturation.",
        "format": "integer",
    },
    {
        "name": "num_rpu_enriched_genes_off_target",
        "display_name": "Number of Enriched Non-Targeted Genes",
        "description": "Genes are classified with a two-class Gaussian mixture model into two groups based on the mean Reads per UMI per gene. Enriched genes have higher mean Reads Per UMI, indicating they have been sequenced to much higher saturation.",
        "format": "integer",
    },
    {
        "name": "mean_reads_per_umi_per_gene_on_target",
        "display_name": "Mean Reads per UMI per Targeted Gene",
        "description": "Mean number of reads per UMI per targeted gene, averaged across all genes.",
        "format": "%.2f",
    },
    {
        "name": "mean_reads_per_umi_per_gene_off_target",
        "display_name": "Mean Reads per UMI per Non-Targeted Gene",
        "description": "Mean number of reads per UMI per non-targeted gene, averaged across all genes.",
        "format": "%.2f",
    },
    {
        "name": "filtered_target_umi_reads_frac",
        "display_name": "Fraction of Reads Removed by UMI Filtering",
        "description": "Fraction of Reads confidently mapped to the Targeted Transcriptome and removed by targeted UMI filtering.",
        "format": "%.2f",
    },
    {
        "name": "filtered_target_umi_count_threshold",
        "display_name": "Reads per UMI threshold for UMI filtering",
        "description": "Reads per UMI threshold for UMI filtering. UMIs in targeted genes where the read support is (strictly) lower than this threshold are filtered out.",
        "format": "integer",
    },
]

# Note: these metrics only apply when batch correction is enabled
CHEMISTRY_BATCH_CORRECTION_METRICS = [
    {
        "name": "batch_effect_score_before_correction",
        "display_name": "Batch Effect Score before Correction",
        "description": "Before chemistry batch correction, the average fraction of 100 nearest neighbors belonging to the same batch, normalized by the expected value without batch effect.",
        "format": "%.2f",
    },
    {
        "name": "batch_effect_score_after_correction",
        "display_name": "Batch Effect Score after Correction",
        "description": "After chemistry batch correction, the average fraction of 100 nearest neighbors belonging to the same batch, normalized by the expected value without batch effect.",
        "format": "%.2f",
    },
]

METRICS = [
    {"name": "Summary", "metrics": SUMMARY_METRICS},
    {"name": "Sequencing", "metrics": SEQUENCING_METRICS},
    {"name": "Mapping", "metrics": MAPPING_METRICS},
    {"name": "Cells", "metrics": DETECTED_CELL_METRICS},
    {"name": "CRISPR Sequencing", "metrics": CRISPR_SEQUENCING_METRICS},
    {"name": "CRISPR Application", "metrics": CRISPR_APPLICATION_METRICS},
    {"name": "Multiplexing Sequencing", "metrics": MULTIPLEXING_SEQUENCING_METRICS},
    {"name": "Multiplexing Application", "metrics": MULTIPLEXING_APPLICATION_METRICS},
    {"name": "GEM Partitions", "metrics": GEM_METRICS},
    {"name": "Aggregation", "metrics": AGGREGATION_METRICS},
    {"name": "CRISPR Aggregation", "metrics": CRISPR_AGGREGATION_METRICS},
    {"name": "Antibody Aggregation", "metrics": ANTIBODY_AGGREGATION_METRICS},
    {"name": "Feature Aggregation", "metrics": CUSTOM_FEATURE_AGGREGATION_METRICS},
    {"name": "Antibody Cell Calling", "metrics": ANTIBODY_CELL_CALLING_METRICS},
    {"name": "Antibody Sequencing", "metrics": ANTIBODY_SEQUENCING_METRICS},
    {"name": "Antibody Application", "metrics": ANTIBODY_APPLICATION_METRICS},
    # Non- (CRISPR or Antibody) feature metrics
    {"name": "Feature Sequencing", "metrics": CUSTOM_FEATURE_SEQUENCING_METRICS},
    {"name": "Feature Application", "metrics": CUSTOM_FEATURE_APPLICATION_METRICS},
    {"name": "Chemistry Batch Correction", "metrics": CHEMISTRY_BATCH_CORRECTION_METRICS,},
    {"name": "Targeted Enrichment", "metrics": TARGETED_APPLICATION_METRICS},
]

# pylint: disable=line-too-long
CHARTS = [
    {
        "layout": {
            "title": "Barcode Rank",
            "width": 470,
            "height": 313,
            "margin": {"l": 60, "r": 0, "t": 30, "b": 40},
            "hovermode": "closest",
            "xaxis": {"title": "Barcodes", "type": "log"},
            "yaxis": {"title": "UMI counts", "type": "log"},
        },
        "data": [],
        "config": shared.CHARTS_PLOTLY_FIXED_CONFIG,
        "function": "plot_barcode_rank",
        "description": "The plot shows the count of filtered UMIs mapped to each barcode.  Barcodes can be determined to be cell-associated based on their UMI counts or by their expression profiles. Therefore some regions of the graph can contain both cell-associated and background-associated barcodes. The color of the graph represents the local density of barcodes that are cell-associated.",
        "name": "barcode_rank",
    },
    {
        "layout": {
            "title": "Cell UMI Counts",
            "width": 588,
            "height": 400,
            "margin": {"l": 65, "r": 0, "t": 30, "b": 70},
            "hovermode": "closest",
            "showlegend": True,
        },
        "data": [{"x": [], "y": [], "mode": "markers", "type": "scattergl"}],
        "config": shared.CHARTS_PLOTLY_MOVABLE_CONFIG,
        "description": "Each point represents a cell-barcode. The axes measure the total UMI counts in each cell-barcode that mapped to each transcriptome reference. The points are colored by the number of inferred cells in the GEM associated with each barcode. A multiplet represents either a GEM inferred to have encapsulated >1 cell or a barcode sequence that was shared by multiple single-cell GEMs.",
        "name": "barnyard_counts",
        "function": "plot_barnyard_barcode_counts",
    },
    {
        "layout": {
            "title": "t-SNE Projection of Cells Colored by UMI Counts",
            "width": 588,
            "height": 400,
            "margin": {"l": 70, "r": 0, "t": 30, "b": 70},
            "hovermode": "closest",
            "xaxis": {"title": "t-SNE1"},
            "yaxis": {"title": "t-SNE2"},
        },
        "data": [
            {
                "x": [],
                "y": [],
                "text": [],
                "hoverinfo": "text",
                "mode": "markers",
                "type": "scattergl",
                "marker": {"color": [], "size": 4, "showscale": True, "colorscale": "Jet",},
            }
        ],
        "config": shared.CHARTS_PLOTLY_MOVABLE_CONFIG,
        "name": "tsne_counts",
        "description": "Shown here are the total UMI counts for each cell-barcode. Cells with greater UMI counts likely have higher RNA content than cells with fewer UMI counts. The axes correspond to the 2-dimensional embedding produced by the t-SNE algorithm. In this space, pairs of cells that are close to each other have more similar gene expression profiles than cells that are distant from each other. The display is limited to a random subset of %d cells."
        % MAX_WEBSHIM_BCS_DIM,
        "function": "plot_tsne_totalcounts",
    },
    {
        "layout": {
            "title": "t-SNE projection of Cells Colored by Automated Clustering",
            "width": 588,
            "height": 400,
            "margin": {"l": 70, "r": 0, "t": 30, "b": 70},
            "hovermode": "closest",
            "xaxis": {"title": "t-SNE1"},
            "yaxis": {"title": "t-SNE2"},
        },
        "name": "tsne_clustering",
        "description": "These are the assignments of each cell-barcode a clusters by an automated clustering algorithm. The clustering groups together cells that have similar expression profiles. The axes correspond to the 2-dimensional embedding produced by the t-SNE algorithm. In this space, pairs of cells that are close to each other have more similar gene expression profiles than cells that are distant from each other. The display is limited to a random subset of %d cells and K-means up to K=%d. Please use Loupe Browser to view the entire dataset."
        % (MAX_WEBSHIM_BCS_DIM, MAX_WEBSHIM_KMEANS_K),
        "function": "plot_tsne",
    },
    {
        "table": {"width": 999, "height": 1000},
        "title": "Top Genes By Cluster (Log2 fold-change, p-value)",
        "name": "differential_expression",
        "description": 'The differential expression analysis seeks to find, for each cluster, genes that are more highly expressed in that cluster relative to the rest of the sample. Here a differential expression test was performed between each cluster and the rest of the sample for each gene. The Log2 fold-change (L2FC) is an estimate of the log2 ratio of expression in a cluster to that in all other cells. A value of 1.0 indicates 2-fold greater expression in the cluster of interest. The p-value is a measure of the statistical significance of the expression difference and is based on a negative binomial test. The p-value reported here has been adjusted for multiple testing via the Benjamini-Hochberg procedure. In this table you can click on a column to sort by that value. Also, in this table genes were filtered by (Mean UMI counts > %0.1f) and the top N genes by L2FC for each cluster were retained. Genes with L2FC < 0 or adjusted p-value >= %0.2f were grayed out. The number of top genes shown per cluster, N, is set to limit the number of table entries shown to %d; N=%d/K^2 where K is the number of clusters. N can range from 1 to %d. For the full table, please refer to the "differential_expression.csv" files produced by the pipeline.'
        % (
            TOP_DE_GENES_MIN_MEAN,
            PVALUE_DEEMPHASIS_CUTOFF,
            MAX_DE_TABLE_ENTRIES,
            MAX_DE_TABLE_ENTRIES,
            MAX_TOP_N_GENES,
        ),
        "function": "plot_differential_expression",
    },
    {
        "layout": {
            "title": "Sequencing Saturation",
            "width": 588,
            "height": 400,
            "margin": {"l": 70, "r": 65, "t": 30, "b": 70},
            "showlegend": False,
            "xaxis": {"title": "Mean Reads per Cell"},
            "yaxis": {"title": "Sequencing Saturation", "range": [0, 1]},
            "shapes": [
                {
                    "type": "line",
                    "x0": 0,
                    "y0": SATURATION_LINE,
                    "x1": 0,
                    "y1": SATURATION_LINE,
                    "line": {"color": "rgb(128, 128, 128)", "width": 4, "dash": "dot"},
                }
            ],
        },
        "data": [],  # data entries are built in the function
        "kwargs": {"metric_suffix": "subsampled_duplication_frac", "show_multi_genome_only": True,},
        "config": shared.CHARTS_PLOTLY_MOVABLE_CONFIG,
        "name": "sequencing_saturation",
        "description": "This plot shows the %s metric as a function of downsampled sequencing depth in mean reads per cell, up to the observed sequencing depth. %s is a measure of the observed library complexity, and approaches 1.0 (100%%) when all converted mRNA transcripts have been sequenced. The slope of the curve near the endpoint can be interpreted as an upper bound to the benefit to be gained from increasing the sequencing depth beyond this point. The dotted line is drawn at a value reasonably approximating the saturation point."
        % (
            SEQUENCING_SATURATION_METRIC["display_name"],
            SEQUENCING_SATURATION_METRIC["display_name"],
        ),
        "function": "plot_subsampled_scatterplot_metric",
    },
    {
        "layout": {
            "title": "Median Genes per Cell",
            "width": 588,
            "height": 400,
            "margin": {"l": 70, "r": 0, "t": 30, "b": 70},
            "showlegend": True,
            "xaxis": {"title": "Mean Reads per Cell"},
            "yaxis": {"title": "Median Genes per Cell"},
        },
        "data": [],  # data entries are built in the function
        "kwargs": {"metric_suffix": "subsampled_filtered_bcs_median_unique_genes_detected"},
        "kwargs_prefixes": ["references"],
        "config": shared.CHARTS_PLOTLY_MOVABLE_CONFIG,
        "name": "median_genes_per_cell",
        "description": "This plot shows the %s as a function of downsampled sequencing depth in mean reads per cell, up to the observed sequencing depth. The slope of the curve near the endpoint can be interpreted as an upper bound to the benefit to be gained from increasing the sequencing depth beyond this point."
        % (GENES_PER_DETECTED_CELL_METRIC["display_name"]),
        "function": "plot_subsampled_scatterplot_metric",
    },
]

METRIC_ALARMS = [
    {
        "name": MULTI_NUMBER_OF_DETECTED_CELLS_METRIC["name"],
        "format": MULTI_NUMBER_OF_DETECTED_CELLS_METRIC["format"],
        "error": {
            "title": "No Cells Detected",
            "message": "No valid sequencing data was detected. Please check the sequencing data.",
            "test": "< 1",
        },
        "warn": {
            "title": "Low Number of Cells Detected",
            "message": "Ideal > 100. This usually indicates poor cell handling, poor library quality, or poor sequencing quality. Application performance is likely to be affected.",
            "test": "< 100",
        },
    },
    {
        "name": MULTI_NUMBER_OF_ANTIBODY_DETECTED_CELLS_METRIC["name"],
        "format": MULTI_NUMBER_OF_ANTIBODY_DETECTED_CELLS_METRIC["format"],
        "error": {
            "title": "No Cells Detected",
            "message": "No valid sequencing data was detected. Please check the sequencing data.",
            "test": "< 1",
        },
        "warn": {
            "title": "Low Number of Cells Detected",
            "message": "Ideal > 100. This usually indicates poor cell handling, poor library quality, or poor sequencing quality. Might indicate a need for a larger antibody panel. Application performance is likely to be affected.",
            "test": "< 100",
        },
    },
    {
        "name": GOOD_BCS_METRIC["name"],
        "format": GOOD_BCS_METRIC["format"],
        "error": {
            "title": "Low Fraction Valid Barcodes",
            "message": "Ideal > 75%. This may indicate a quality issue with the Illumina I7 read for Single Cell 3' v1 or the R1 read for Single Cell 3' v2/v3 and Single Cell 5'. Application performance is likely to be affected.",
            "test": "< 0.5",
        },
        "warn": {
            "title": "Low Fraction Valid Barcodes",
            "message": "Ideal > 75%. This may indicate a quality issue with the Illumina I7 read for Single Cell 3' v1 or the R1 read for Single Cell 3' v2/v3 and Single Cell 5'. Application performance may be affected.",
            "test": "< 0.75",
        },
    },
    {
        "name": GOOD_UMIS_METRIC["name"],
        "format": GOOD_UMIS_METRIC["format"],
        "error": {
            "title": "Low Fraction Valid UMIs",
            "message": "Ideal > 75%. This may indicate a quality issue with the Illumina R2 read for Single Cell 3' v1 or the R1 read for Single Cell 3' v2/v3 and Single Cell 5'. Application performance is likely to be affected.",
            "test": "< 0.5",
        },
        "warn": {
            "title": "Low Fraction Valid UMIs",
            "message": "Ideal > 75%. This may indicate a quality issue with the Illumina R2 read for Single Cell 3' v1 or the R1 read for Single Cell 3' v2/v3 and Single Cell 5'. Application performance may be affected.",
            "test": "< 0.75",
        },
    },
    {
        "name": MULTI_TRANSCRIPTOME_CONF_MAPPED_READS_METRIC["name"],
        "format": MULTI_TRANSCRIPTOME_CONF_MAPPED_READS_METRIC["format"],
        "error": {
            "title": "Low Fraction Reads Confidently Mapped To Transcriptome",
            "message": "Ideal > 30%. This can indicate use of the wrong reference transcriptome, a reference transcriptome with overlapping genes, poor library quality, poor sequencing quality, or reads shorter than the recommended minimum. Application performance is likely to be affected.",
            "test": "< 0.20",
        },
        "warn": {
            "title": "Low Fraction Reads Confidently Mapped To Transcriptome",
            "message": "Ideal > 30%. This can indicate use of the wrong reference transcriptome, a reference transcriptome with overlapping genes, poor library quality, poor sequencing quality, or reads shorter than the recommended minimum. Application performance may be affected.",
            "test": "< 0.30",
        },
    },
    {
        "name": ANTISENSE_CONF_MAPPED_READS_METRIC["name"],
        "format": ANTISENSE_CONF_MAPPED_READS_METRIC["format"],
        "error": {
            "title": "High Fraction of Reads Mapped Antisense to Genes",
            "message": "Ideal < 10%. This can indicate use of an unsupported chemistry type (e.g. using Single Cell V(D)J for gene counting). Application performance is likely to be affected.",
            "test": "> 0.30",
        },
        "warn": {
            "title": "High Fraction of Reads Mapped Antisense to Genes",
            "message": "Ideal < 10%. This can indicate use of an unsupported chemistry type (e.g. using Single Cell V(D)J for gene counting). Application performance may be affected.",
            "test": "> 0.10",
        },
    },
    {
        "name": "bc_bases_with_q30_frac",
        "format": "percent",
        "error": {
            "title": "Low Barcode Q30 Fraction (Illumina I7 Read for Single Cell 3' v1, R1 for Single Cell 3' v2/v3 and Single Cell 5')",
            "message": "Ideal > 55%. Application performance is likely to be affected.",
            "test": "< 0.45",
        },
        "warn": {
            "title": "Low Barcode Q30 Fraction (Illumina I7 Read for Single Cell 3' v1, R1 for Single Cell 3' v2/v3 and Single Cell 5')",
            "message": "Ideal > 55%. Application performance may be affected.",
            "test": "< 0.55",
        },
    },
    {
        "name": "read_bases_with_q30_frac",
        "format": "percent",
        "error": {
            "title": "Low RNA Read Q30 Fraction (Illumina R1 for Single Cell 3' v1 and Single Cell 5' paired end, R2 for Single Cell 3' v2/v3 and Single Cell 5' R2-only)",
            "message": "Ideal > 65%. Application performance is likely to be affected.",
            "test": "< 0.55",
        },
        "warn": {
            "title": "Low RNA Read Q30 Fraction (Illumina R1 for Single Cell 3' v1 and Single Cell 5' paired end, R2 for Single Cell 3' v2/v3 and Single Cell 5' R2-only)",
            "message": "Ideal > 65%. Application performance may be affected.",
            "test": "< 0.65",
        },
    },
    {
        "name": "umi_bases_with_q30_frac",
        "format": "percent",
        "error": {
            "title": "Low UMI Q30 Fraction (Illumina R2 Read for Single Cell 3' v1, R1 for Single Cell 3' v2/v3 and Single Cell 5')",
            "message": "Ideal > 75%. Application performance is likely to be affected.",
            "test": "< 0.65",
        },
        "warn": {
            "title": "Low UMI Q30 Fraction (Illumina R2 Read for Single Cell 3' v1, R1 for Single Cell 3' v2/v3 and Single Cell 5')",
            "message": "Ideal > 75%. Application performance may be affected.",
            "test": "< 0.75",
        },
    },
    {
        "name": MULTI_FRAC_READS_IN_CELLS_METRIC["name"],
        "format": MULTI_FRAC_READS_IN_CELLS_METRIC["format"],
        "error": {
            "title": "Low Fraction Reads in Cells",
            "message": "Ideal > 70%. Application performance is likely to be affected. Many of the reads were not assigned to cell-associated barcodes. This could be caused by high levels of ambient RNA or by a significant population of cells with a low RNA content, which the algorithm did not call as cells. The latter case can be addressed by inspecting the data to determine the appropriate cell count and using --force-cells.",
            "test": "< 0.50",
        },
        "warn": {
            "title": "Low Fraction Reads in Cells",
            "message": "Ideal > 70%. Application performance may be affected. Many of the reads were not assigned to cell-associated barcodes. This could be caused by high levels of ambient RNA or by a significant population of cells with a low RNA content, which the algorithm did not call as cells. The latter case can be addressed by inspecting the data to determine the appropriate cell count and using --force-cells.",
            "test": "< 0.70",
        },
    },
    # Batch Aggregation-specific alerts
    {
        "name": "lowest_frac_reads_kept",
        "format": "percent",
        "error": {
            "title": "Low Post-Normalization Read Depth",
            "message": "Ideal > 50%. There may be large differences in sequencing depth across the input libraries. Application performance is likely to be affected.",
            "test": "< 0.25",
        },
        "warn": {
            "title": "Low Post-Normalization Read Depth",
            "message": "Ideal > 50%. There may be large differences in sequencing depth across the input libraries. Application performance may be affected.",
            "test": "< 0.50",
        },
    },
    # Antibody aggregation-specific alerts
    {
        "name": REPORT_PREFIX_ANTIBODY + "reads_lost_to_aggregate_GEMs",
        "format": "percent",
        "error": {
            "title": "High fraction of Antibody library reads coming from aggregate barcodes",
            "message": "Ideal < 5%. Detected some aggregate barcodes. These barcodes and their corresponding reads have been removed from the final matrix.",
            "test": "> 0.5",
        },
        "warn": {
            "title": "High fraction of Antibody library reads coming from aggregate barcodes",
            "message": "Ideal < 5%. Detected some aggregate barcodes. These barcodes and their corresponding reads have been removed from the final matrix.",
            "test": "> 0.05",
        },
    },
    # Targeting-specific alerts
    {
        "name": "multi_frac_conf_transcriptomic_reads_on_target",
        "format": "percent",
        "error": {
            "title": "Low Fraction Reads Confidently Mapped To Targeted Transcriptome",
            "message": "Ideal > 30%. This can indicate use of a small panel with low aggregate expression, use of the wrong target panel, use of a custom panel with high background, or inefficient targeting. Targeted performance may be affected.",
            "test": "> 0.20",
        },
        "warn": {
            "title": "Low Fraction Reads Confidently Mapped To Targeted Transcriptome",
            "message": "Ideal > 30%. This can indicate use of a small panel with low aggregate expression, use of the wrong target panel, use of a custom panel with high background, or inefficient targeting. Targeted performance may be affected.",
            "test": "> 0.30",
        },
    },
    {
        "name": "frac_on_target_genes_enriched",
        "format": "percent",
        "error": {
            "title": "Low Fraction of Targeted Genes Enriched",
            "message": "Ideal > 70%. This can indicate use of the wrong target panel, or inefficient targeting. Targeted performance may be affected.",
            "test": "> 0.70",
        },
        "warn": {
            "title": "Low Fraction of Targeted Genes Enriched",
            "message": "Ideal > 70%. This can indicate use of the wrong target panel, or inefficient targeting. Targeted performance may be affected.",
            "test": "> 0.50",
        },
    },
    {
        "name": "frac_off_target_genes_enriched",
        "format": "percent",
        "error": {
            "title": "High Fraction of Non-Targeted Genes Enriched",
            "message": "This can indicate use of the wrong target panel, or inefficient targeting. Targeted performance may be affected.",
            "test": "< 0.30",
        },
        "warn": {
            "title": "High Fraction of Non-Targeted Genes Enriched",
            "message": "This can indicate use of the wrong target panel, or inefficient targeting. Targeted performance may be affected.",
            "test": "< 0.50",
        },
    },
]

# Targeting color scheme for websummary
ON_TARGET_COLOR = "#0071D9"
OFF_TARGET_COLOR = "darkgrey"

# Targeting column name constants in feature_metrics table
TARGETING_COLNAME = "is_targeted"
READS_COLNAME = "num_reads"
UMIS_COLNAME = "num_umis"
READS_IN_CELLS_COLNAME = "num_reads_cells"
UMIS_IN_CELLS_COLNAME = "num_umis_cells"
LOG_RPU_CELLS_COLNAME = "mean_reads_per_umi_cells_log10"
RPU_COLNAME = "mean_reads_per_umi"
ENRICHMENT_COLNAME = "enriched_rpu"
FEATURE_ID_COLNAME = "feature_id"
FEATURE_NAME_COLNAME = "feature_name"
TARGET_FEATURE_METRICS_COLS = [
    FEATURE_ID_COLNAME,
    FEATURE_NAME_COLNAME,
    TARGETING_COLNAME,
    READS_IN_CELLS_COLNAME,
    UMIS_IN_CELLS_COLNAME,
    LOG_RPU_CELLS_COLNAME,
    RPU_COLNAME,
    ENRICHMENT_COLNAME,
]



#!/usr/bin/env python
#
# Copyright (c) 2018 10X Genomics, Inc. All rights reserved.
#

######################################################
# DO NOT add new items to this file.
#
# - If a constant is only used from a single module, put it in that module.
# - If a constant is only used in association with a particular module, put it
#   in that module.
# - If a constant is used in a bunch of places, create a module with a more
#   specifically descriptive name to put those constants.
######################################################

# tsne
TSNE_N_COMPONENTS = 2
TSNE_DEFAULT_KEY = b"2"
TSNE_DEFAULT_PERPLEXITY = 30
TSNE_THETA = 0.5
RANDOM_STATE = 0
TSNE_MAX_ITER = 1000
TSNE_STOP_LYING_ITER = 250
TSNE_MOM_SWITCH_ITER = 250
ANALYSIS_H5_TSNE_GROUP = "tsne"

# umap
UMAP_N_COMPONENTS = 2
UMAP_DEFAULT_N_NEIGHBORS = 30
UMAP_MIN_DIST = 0.3
UMAP_DEFAULT_METRIC = "correlation"
ANALYSIS_H5_UMAP_GROUP = "umap"

# clustering
ANALYSIS_H5_CLUSTERING_GROUP = "clustering"
ANALYSIS_H5_KMEANS_GROUP = "kmeans"  # Deprecated
ANALYSIS_H5_KMEDOIDS_GROUP = "kmedoids"  # Deprecated
KMEDOIDS_SQMAT_ENTERIES_PER_MEM_GB = 12e7
GRAPHCLUST_NEIGHBORS_DEFAULT = 1
GRAPHCLUST_NEIGHBOR_A_DEFAULT = -230.0
GRAPHCLUST_NEIGHBOR_B_DEFAULT = 120.0

# multigenome
GEM_CLASS_MULTIPLET = b"Multiplet"
GEM_CLASS_GENOME0 = b"genome0"
GEM_CLASS_GENOME1 = b"genome1"
GEM_CLASSES = [GEM_CLASS_GENOME0, GEM_CLASS_GENOME1, GEM_CLASS_MULTIPLET]
DEFAULT_MULTIPLET_THRESHOLD = 10
MULTIPLET_PROB_THRESHOLD = 0.10
COUNT_PURITY_OUTLIER_PROB_THRESHOLD = 0.01

# pca
PCA_N_COMPONENTS_DEFAULT = 10
NUM_IRLB_MATRIX_ENTRIES_PER_MEM_GB = 10e6  # based on empirical testing
IRLB_BASE_MEM_GB = 1
ANALYSIS_H5_PCA_GROUP = "pca"

# chemistry batch correction
# this upper limit was determined via testing,
#   larger numbers of cells will require _at least_ memory reservation changes
#   in CORRECT_CHEMISTRY_BATCH.join, if not substantial algorithmic changes
CBC_MAX_NCELLS = 128000
CBC_N_COMPONENTS_DEFAULT = 100
CBC_KNN = 10
CBC_ALPHA = 0.1
CBC_SIGMA = 150
CBC_REALIGN_PANORAMA = False

# lsa
LSA_N_COMPONENTS_DEFAULT = 15
ANALYSIS_H5_LSA_GROUP = "lsa"

# plsa
PLSA_N_COMPONENTS_DEFAULT = 10
ANALYSIS_H5_PLSA_GROUP = "plsa"
NUM_PLSA_EM_MATRIX_ENTRIES_PER_MEM_GB = 25e6
PLSA_EM_BASE_MEM_GB = 1

# single genome
MIN_N_CLUSTERS = 2
MAX_N_CLUSTERS_DEFAULT = 10
MAX_N_CLUSTERS_ATAC = 5
ANALYSIS_H5_MATRIX_GROUP = "matrix"
ANALYSIS_H5_KMEANS_DIFFERENTIAL_EXPRESSION_GROUP = "differential_expression"  # Deprecated
ANALYSIS_H5_KMEDOIDS_DIFFERENTIAL_EXPRESSION_GROUP = "differential_expression"  # Deprecated
ANALYSIS_H5_DIFFERENTIAL_EXPRESSION_GROUP = "all_differential_expression"
ANALYSIS_H5_DIFFERENTIAL_EXPRESSION_FEATURE_LIST = "diffexp_feature_indices"

ANALYSIS_H5_MAP_CLUSTERING = {
    "pca": ANALYSIS_H5_KMEANS_GROUP,
    "lsa": ANALYSIS_H5_KMEANS_GROUP,
    "plsa": ANALYSIS_H5_KMEANS_GROUP,
}

ANALYSIS_H5_MAP_DE = {
    "pca": ANALYSIS_H5_KMEANS_DIFFERENTIAL_EXPRESSION_GROUP,
    "lsa": ANALYSIS_H5_KMEDOIDS_DIFFERENTIAL_EXPRESSION_GROUP,
    "plsa": ANALYSIS_H5_KMEDOIDS_DIFFERENTIAL_EXPRESSION_GROUP,
}



#!/usr/bin/env python
#
# Copyright (c) 2018 10X Genomics, Inc. All rights reserved.
#

MATRIX_MEM_GB_MULTIPLIER = 2  # TODO reduce this once we're confident about the actual memory bounds
NUM_MATRIX_ENTRIES_PER_MEM_GB = 50e6

# Empirical obs: with new CountMatrix setup, take ~ 50 bytes/bc
NUM_MATRIX_BARCODES_PER_MEM_GB = 10000000
MIN_MEM_GB = 3
MIN_MEM_GB_NOWHITELIST = 64

GZIP_SUFFIX = b".gz"
LZ4_SUFFIX = b".lz4"

H5_COMPRESSION_LEVEL = 1
H5_FILETYPE_KEY = "filetype"
H5_FEATURE_REF_ATTR = "features"
H5_TARGET_SET_ATTR = b"target_sets"
H5_BCS_ATTR = "barcodes"
H5_MATRIX_DATA_ATTR = "data"
H5_MATRIX_INDICES_ATTR = "indices"
H5_MATRIX_INDPTR_ATTR = "indptr"
H5_MATRIX_SHAPE_ATTR = "shape"
H5_MATRIX_ATTRS = {
    H5_MATRIX_DATA_ATTR: "int32",
    H5_MATRIX_INDICES_ATTR: "int64",
    H5_MATRIX_INDPTR_ATTR: "int64",
    H5_MATRIX_SHAPE_ATTR: "int32",
}
H5_CHEMISTRY_DESC_KEY = "chemistry_description"
H5_LIBRARY_ID_MAPPING_KEY = "library_ids"
H5_ORIG_GEM_GROUP_MAPPING_KEY = "original_gem_groups"
H5_METADATA_ATTRS = [
    H5_LIBRARY_ID_MAPPING_KEY,
    H5_ORIG_GEM_GROUP_MAPPING_KEY,
    H5_CHEMISTRY_DESC_KEY,
]



#!/usr/bin/env python
#
# Copyright (c) 2015 10X Genomics, Inc. All rights reserved.
#

"""
Assorted grab-bag of miscellaneous helper methods.

Do not add to this module.  Instead, find or create a module with a name
that indicates to a potential user what sorts of methods they might find
in that module.
"""

from __future__ import absolute_import, annotations, division

import collections
import csv
import itertools
import json
import os
import random
import h5py
import numpy as np
from six import ensure_binary, ensure_text
import bmkcc.tenkit.log_subprocess as tk_subproc
import bmkcc.tenkit.seq as tk_seq
import bmkcc.tenkit.constants as tk_constants
import bmkcc.cellranger.constants as cr_constants
import bmkcc.cellranger.h5_constants as h5_constants

from typing import (  # pylint: disable=unused-import
    AnyStr,
    Dict,
    Generator,
    IO,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from pysam import AlignmentFile, AlignedSegment  # pylint: disable=unused-import


def get_gem_group_from_barcode(barcode: Optional[bytes]) -> Optional[int]:
    """
    Method to get just the gem group from a barcode

    Args:
        barcode (bytes): a barcode, a la "ACGTACTAGAC-1"

    Returns:
        The Gem group as an int or None if not present.
    """
    if barcode is None:
        return None
    gg = barcode.index(b"-")
    if gg > 0:
        return int(barcode[gg + 1 :])
    else:
        return None


def _load_reference_metadata_file(reference_path: str,) -> Dict[str, Union[dict, list, str, int]]:
    reference_metadata_file = os.path.join(reference_path, cr_constants.REFERENCE_METADATA_FILE)
    with open(reference_metadata_file, "r") as f:
        return json.load(f)


def get_reference_star_path(reference_path: str) -> str:
    return os.path.join(reference_path, cr_constants.REFERENCE_STAR_PATH)


def get_reference_genome_fasta(reference_path: str) -> str:
    return os.path.join(reference_path, cr_constants.REFERENCE_FASTA_PATH)


def get_reference_genomes(reference_path: str) -> List[str]:
    data = _load_reference_metadata_file(reference_path)
    return data[cr_constants.REFERENCE_GENOMES_KEY]


def is_arc_reference(reference_path: str) -> bool:
    data = _load_reference_metadata_file(reference_path)
    return data.get("mkref_version", "").startswith("cellranger-arc")


def get_reference_mem_gb_request(reference_path: str) -> int:
    data = _load_reference_metadata_file(reference_path)
    return data[cr_constants.REFERENCE_MEM_GB_KEY]


def barcode_sort_key(read, squash_unbarcoded: bool = False) -> Optional[tuple]:
    formatted_bc = get_read_barcode(read)
    if squash_unbarcoded and formatted_bc is None:
        return None
    (bc, gg) = split_barcode_seq(formatted_bc)
    library_idx = get_read_library_index(read)
    return gg, bc, library_idx, get_read_raw_umi(read)


def pos_sort_key(read: AlignedSegment) -> tuple:
    return read.tid, read.pos


def format_barcode_seq(barcode: bytes, gem_group: Optional[int] = None) -> bytes:
    if gem_group is not None:
        barcode += b"-%d" % gem_group
    return barcode


def format_barcode_seqs(
    barcode_seqs: List[bytes], gem_groups: Optional[Iterable[int]]
) -> List[bytes]:
    if gem_groups is None:
        return barcode_seqs
    new_barcode_seqs = []
    unique_gem_groups = sorted(set(gem_groups))
    for gg in unique_gem_groups:
        new_barcode_seqs += [format_barcode_seq(bc, gg) for bc in barcode_seqs]
    return new_barcode_seqs


def split_barcode_seq(
    barcode: Optional[bytes],
) -> Union[Tuple[None, None], Tuple[bytes, Optional[int]]]:
    if barcode is None:
        return None, None

    barcode_parts = barcode.split(b"-")

    barcode = barcode_parts[0]
    if len(barcode_parts) > 1:
        gem_group = int(barcode_parts[1])
    else:
        gem_group = None

    return barcode, gem_group


def bcs_suffices_to_names(bcs: Iterable[bytes], gg_id_map: Dict[int, AnyStr]) -> List[AnyStr]:
    """Turn list of barcodes into corresponding list of mapped values.

    Args:
        bcs (list): aggr barcodes with suffix corresponding to gem group id
        gg_id_map (dict): mapping gem-group-ids/barcode-suffixes (int) to
            a desired named value (typically `aggr_id` for the library id of that
            one sample, or `batch_id` for the name of a batch it is a part of.)
    """
    mapped_ids = [gg_id_map[split_barcode_seq(bc)[1]] for bc in bcs]
    return mapped_ids


def is_barcode_corrected(raw_bc_seq: bytes, processed_bc_seq: Optional[bytes]) -> bool:
    if processed_bc_seq is None:
        return False

    bc_seq, _ = split_barcode_seq(processed_bc_seq)
    return bc_seq != raw_bc_seq


def get_genome_from_str(s, genomes):
    assert len(genomes) > 0

    if s is None:
        return None

    if len(genomes) == 1:
        return genomes[0]

    for genome in genomes:
        if s.startswith(genome):
            return genome

    raise Exception("%s does not have valid associated genome" % s)


def remove_genome_from_str(s, genomes, prefixes_as_genomes=False):
    assert len(genomes) > 0
    if s is None:
        return None
    if len(genomes) == 1:
        return s

    # Gene names/ids/chroms are padded with N underscores to achieve the same prefix length
    #   for all genomes, e.g., GRCh38_* and mm10___*
    max_len = max(len(g) for g in genomes)

    for genome in genomes:
        if s.startswith(genome):
            # Strip genome and subsequent underscores
            if prefixes_as_genomes:
                return s[(1 + len(genome)) :]
            else:
                return s[(1 + max_len) :]

    raise Exception("%s does not have valid associated genome" % s)


def get_genome_from_read(read, chroms, genomes):
    assert len(genomes) > 0

    if read.is_unmapped:
        return None

    if len(genomes) == 1:
        return genomes[0]

    return get_genome_from_str(chroms[read.tid], genomes)


def get_read_barcode(read: AlignedSegment):
    return _get_read_tag(read, cr_constants.PROCESSED_BARCODE_TAG)


def get_read_raw_barcode(read: AlignedSegment):
    return _get_read_tag(read, cr_constants.RAW_BARCODE_TAG)


def get_read_barcode_qual(read: AlignedSegment):
    return _get_read_tag(read, cr_constants.RAW_BARCODE_QUAL_TAG)


def get_read_umi_qual(read: AlignedSegment):
    return _get_read_tag(read, cr_constants.UMI_QUAL_TAG)


def get_read_umi(read: AlignedSegment):
    return _get_read_tag(read, cr_constants.PROCESSED_UMI_TAG)


def get_read_raw_umi(read: AlignedSegment):
    return _get_read_tag(read, cr_constants.RAW_UMI_TAG)


def get_read_library_index(read: AlignedSegment):
    return _get_read_tag(read, cr_constants.LIBRARY_INDEX_TAG)


def get_read_gene_ids(read: AlignedSegment):
    s = _get_read_tag(read, cr_constants.FEATURE_IDS_TAG)  # type: Optional[str]
    if s is None:
        return None
    assert isinstance(s, str)
    return tuple(s.split(";"))


def get_read_transcripts_iter(read: AlignedSegment):
    """Iterate over all transcripts compatible with the given read.

    We do this by iterating over the TX tag entries that are of the form
    `TX:<transcript id>,<strand><position>,<cigar>`.

    Note:
        When intronic alignment is turned on, the `TX` tag is set to
        `TX:<gene id>,<strand>` for intronic  reads or exonic reads that are
        not compatible with annotated splice junctions. We ignore these.
    """
    s = _get_read_tag(read, cr_constants.TRANSCRIPTS_TAG)
    if s is None:
        return

    s = ensure_binary(s)
    for x in s.split(b";"):
        if len(x) == 0:
            continue

        parts = x.split(b",")
        # len(parts) = 3 for transcriptomic alignments, while for intronic alignments when
        # intron counting mode is enabled len(parts)=2
        if len(parts) != 3:
            continue

        chrom = parts[0] if parts[0] else None
        if parts[1]:
            strand = parts[1][0:1]
            pos = int(parts[1][1:])
        else:
            strand = None
            pos = None
        cigarstring = parts[2] if parts[2] else None
        assert strand in cr_constants.STRANDS or strand is None

        yield chrom, strand, pos, cigarstring


REGION_TAG_MAP = {
    "E": cr_constants.EXONIC_REGION,
    "N": cr_constants.INTRONIC_REGION,
    "I": cr_constants.INTERGENIC_REGION,
}


def get_mapping_region(read: AlignedSegment) -> Optional[str]:
    region_tag = _get_read_tag(read, "RE")
    return REGION_TAG_MAP.get(region_tag, None)


def iter_by_qname(in_genome_bam, in_trimmed_bam):
    """Iterate through multiple BAMs by qname simultaneously.

    Assume the trimmed-read-bam has every qname in the genome bam, in the same order.
    """

    genome_bam_iter = itertools.groupby(in_genome_bam, key=lambda read: read.qname)

    if in_trimmed_bam is None:
        trimmed_bam_iter = iter(())
    else:
        trimmed_bam_iter = itertools.groupby(in_trimmed_bam, key=lambda read: read.qname)

    for (genome_qname, genome_reads), trimmed_tuple in itertools.zip_longest(
        genome_bam_iter, trimmed_bam_iter
    ):
        trimmed_qname, trimmed_reads = trimmed_tuple or (None, [])
        genome_reads = list(genome_reads)
        trimmed_reads = list(trimmed_reads)

        assert (in_trimmed_bam is None) or trimmed_qname == genome_qname
        yield (genome_qname, genome_reads, trimmed_reads)


def set_read_tags(read: AlignedSegment, tags) -> None:
    read_tags = read.tags
    read_tags.extend(tags)
    read.tags = read_tags


def _get_read_tag(read: AlignedSegment, tag):
    try:
        r = read.opt(tag)
        if not r:
            r = None
        return r
    except KeyError:
        return None


def get_fastq_read1(r1, r2, reads_interleaved):
    if reads_interleaved:
        name1, seq1, qual1, _, _, _ = r1
    else:
        name1, seq1, qual1 = r1
    return name1, seq1, qual1


def get_fastq_read2(r1, r2, reads_interleaved):
    if reads_interleaved:
        _, _, _, name2, seq2, qual2 = r1
    else:
        name2, seq2, qual2 = r2
    return name2, seq2, qual2


def get_read_extra_flags(read: AlignedSegment):
    return _get_read_tag(read, cr_constants.EXTRA_FLAGS_TAG) or 0


# EXTRA_FLAGS_CONF_MAPPED_TXOME = 1
EXTRA_FLAGS_LOW_SUPPORT_UMI = 2
# EXTRA_FLAGS_GENE_DISCORDANT = 4
EXTRA_FLAGS_UMI_COUNT = 8
EXTRA_FLAGS_CONF_MAPPED_FEATURE = 16
EXTRA_FLAGS_FILTERED_TARGET_UMI = 32


def is_read_low_support_umi(read: AlignedSegment) -> bool:
    return (get_read_extra_flags(read) & EXTRA_FLAGS_LOW_SUPPORT_UMI) > 0


def is_read_filtered_target_umi(read: AlignedSegment) -> bool:
    return (get_read_extra_flags(read) & EXTRA_FLAGS_FILTERED_TARGET_UMI) > 0


def is_read_umi_count(read: AlignedSegment) -> bool:
    return (get_read_extra_flags(read) & EXTRA_FLAGS_UMI_COUNT) > 0


def is_read_conf_mapped_to_feature(read: AlignedSegment) -> bool:
    return (get_read_extra_flags(read) & EXTRA_FLAGS_CONF_MAPPED_FEATURE) > 0


def is_read_dupe_candidate(read, high_conf_mapq, use_corrected_umi=True, use_umis=True) -> bool:
    if use_corrected_umi:
        umi = get_read_umi(read)
    else:
        umi = get_read_raw_umi(read)

    return (
        not read.is_secondary
        and (umi is not None or not use_umis)
        and (get_read_barcode(read) is not None)
        and not is_read_low_support_umi(read)
        and not is_read_filtered_target_umi(read)
        and (
            is_read_conf_mapped_to_transcriptome(read, high_conf_mapq)
            or is_read_conf_mapped_to_feature(read)
        )
    )


def is_read_conf_mapped(read: AlignedSegment, high_conf_mapq) -> bool:
    if read.is_unmapped:
        return False
    elif read.mapq < high_conf_mapq:
        return False
    return True


def is_read_conf_mapped_to_transcriptome(read: AlignedSegment, high_conf_mapq) -> bool:
    if is_read_conf_mapped(read, high_conf_mapq):
        gene_ids = get_read_gene_ids(read)
        return (gene_ids is not None) and (len(gene_ids) == 1)
    return False


def is_read_conf_mapped_to_transcriptome_barcoded(read, high_conf_mapq) -> bool:
    return (
        is_read_conf_mapped_to_transcriptome(read, high_conf_mapq)
        and not read.is_secondary
        and get_read_barcode(read)
    )


def is_read_conf_mapped_to_transcriptome_barcoded_deduped(read, high_conf_mapq) -> bool:
    return (
        is_read_conf_mapped_to_transcriptome_barcoded(read, high_conf_mapq)
        and not read.is_duplicate
    )


def is_read_conf_mapped_to_transcriptome_barcoded_deduped_with_umi(read, high_conf_mapq) -> bool:
    return is_read_conf_mapped_to_transcriptome_barcoded_deduped(
        read, high_conf_mapq
    ) and get_read_umi(read)


def load_barcode_summary(barcode_summary):
    if barcode_summary:
        with h5py.File(barcode_summary) as f:
            return list(f[cr_constants.H5_BC_SEQUENCE_COL])
    return None


def compress_seq(s: bytes, bits: int = 64):
    """Pack a DNA sequence (no Ns!) into a 2-bit format, into a python int.

    Args:
        s (str): A DNA sequence.
    Returns:
        int: The sequence packed into the bits of an integer.
    """
    assert len(s) <= (bits // 2 - 1)
    result = 0
    for i in range(len(s)):
        nuc = s[i : i + 1]
        assert nuc in tk_seq.NUCS_INVERSE
        result = result << 2
        result = result | tk_seq.NUCS_INVERSE[nuc]
    return result


def set_tag(read, key, old_value, new_value):
    """ Set a bam tag for a read, overwriting the previous value """
    tags = read.tags
    tags.remove((key, old_value))
    tags.append((key, new_value))
    read.tags = tags


def build_alignment_param_metrics() -> Dict[str, Union[str, int]]:
    """ hard-coded aligner settings"""

    align = {"aligner": "star", "high_conf_mapq": cr_constants.STAR_DEFAULT_HIGH_CONF_MAPQ}
    aln_param_metrics = {}
    for key, value in align.items():
        aln_param_metrics["alignment_" + key] = value
    return aln_param_metrics


def get_high_conf_mapq() -> int:
    return cr_constants.STAR_DEFAULT_HIGH_CONF_MAPQ


def get_mem_gb_request_from_genome_fasta(reference_path: str) -> float:
    in_fasta_fn = get_reference_genome_fasta(reference_path)
    genome_size_gb = float(os.path.getsize(in_fasta_fn)) / 1e9
    return np.ceil(
        max(
            h5_constants.MIN_MEM_GB,
            cr_constants.BAM_CHUNK_SIZE_GB + max(1, 2 * int(genome_size_gb)),
        )
    )


def merge_jsons_as_dict(in_filenames: Iterable[Union[str, bytes, os.PathLike]]) -> dict:
    """ Merge a list of json files and return the result as a dictionary """
    d = {}
    for filename in in_filenames:
        if (filename is None) or (not (os.path.isfile(filename))):
            continue
        try:
            with open(filename, "r") as f:
                data = json.load(f)
                d.update(data)
        except IOError:
            continue
    return d


def format_barcode_summary_h5_key(
    library_prefix: str, genome: str, region: str, read_type: str
) -> str:
    """Formats the barcode summary into a key to access the `barcode_summary.h5` outs.

    Here, we need to accommodate both accessing old and new versions of the h5 file
    compatible with both v2 and v3 of the matrix.

    Args:
        library_prefix (str): Name of the library, used as a prefix
        genome         (str): Name of the genome used for alignment
        region         (str): Name of the subset of the genome we are looking at (e.g. transcriptome, regulome,
                            epigenome, ...). This should be a controlled vocabulary in `cellranger.constants`
        read_type      (str): Name of the read types we are trying to extract (e.g. `conf_mapped_deduped_barcoded`, ...).
                            It should be also controlled in `cellranger.constants`.
    Returns:
        output_key     (str): A string constant with the suffix `reads` appended.
    """
    # Clean the input for non-found or undefined keys. Will break badly if they are not controlled, but will let us
    # keep compatible across v2 and v3 developments.
    str_list = [library_prefix, genome, region, read_type]
    # Append reads
    str_list.append("reads")
    # join
    output_key = "_".join(str_list)
    return output_key


def downsample(rate: Optional[float]) -> bool:
    if rate is None or rate == 1.0:
        return True
    return random.random() <= rate


def numpy_groupby(values, keys) -> Generator[Tuple[tuple, tuple], None, None]:
    """Group a collection of numpy arrays by key arrays.

    Yields `(key_tuple, view_tuple)` where `key_tuple` is the key grouped
    on and `view_tuple` is a tuple of views into the value arrays.

    Args:
        values (tuple of arrays): tuple of arrays to group.
        keys (tuple): tuple of sorted, numeric arrays to group by.
    Returns:
        sequence of tuple: Sequence of (`key_tuple`, `view_tuple`).
    """

    if len(values) == 0:
        return
    if len(values[0]) == 0:
        return

    for key_array in keys:
        assert len(key_array) == len(keys[0])
    for value_array in values:
        assert len(value_array) == len(keys[0])

    # The indices where any of the keys differ from the previous key become group boundaries
    # pylint: disable=no-member
    key_change_indices = np.logical_or.reduce(
        tuple(np.concatenate(([1], np.diff(key))) != 0 for key in keys)
    )
    group_starts = np.flatnonzero(key_change_indices)
    group_ends = np.roll(group_starts, -1)
    group_ends[-1] = len(keys[0])

    for group_start, group_end in zip(group_starts, group_ends):
        yield tuple(key[group_start] for key in keys), tuple(
            value[group_start:group_end] for value in values
        )


def get_seqs(l: int) -> List[bytes]:
    if l == 1:
        return tk_seq.NUCS
    old_seqs = get_seqs(l - 1)
    new_seqs = []  # type: List[bytes]
    for old_seq in old_seqs:
        for base in tk_seq.NUCS:
            new_seqs.append(old_seq + base)
    return new_seqs


def get_fasta_iter(f: IO[bytes]) -> Generator[Tuple[bytes, bytes], None, None]:
    hdr = b""
    seq = b""
    for line in f:
        line = line.strip()
        if line.startswith(b">"):
            if hdr:
                yield hdr, seq
            hdr = line[1:]
            seq = b""
        else:
            seq += line
    if hdr:
        yield hdr, seq


# Pysam numeric codes to meaningful categories
_cigar_numeric_to_category_map = {
    0: "M",
    1: "I",
    2: "D",
    3: "N",
    4: "S",
    5: "H",
    6: "P",
    7: "=",
    8: "X",
}


def get_cigar_summary_stats(read: AlignedSegment, strand: bytes) -> Dict[str, int]:
    """
    Get number of mismatches, insertions, deletions, ref skip, soft clip, hard clip bases from a read.

    Returns a dictionary by the element's CIGAR designation. Adds additional
    fields to distinguish between three and five prime soft-clipping for `R1`
    and `R2`: `R1_S_three_prime` and `R1_S_five_prime`, etc. to account for
    soft-clipped local alignments.

    Args:
        read (pysam.AlignedRead): aligned read object
        strand (string): + or - to indicate library orientation (MRO argument strand, for example)

    Returns:
        dict of str,int: Key of base type to base counts for metrics. Adds
                         additional fields to distinguish between three and
                         five prime soft-clipping: `S_three_prime` and
                         `S_five_prime`.
    """

    statistics = {}  # Dict[str, int]
    cigar_tuples = read.cigar

    for i, (category, count) in enumerate(cigar_tuples):
        # Convert numeric code to category
        category = _cigar_numeric_to_category_map[category]
        count = int(count)

        # Detect 5 prime soft-clipping
        if i == 0:

            if strand == cr_constants.REVERSE_STRAND:
                metric = "R1_S_five_prime" if read.is_read1 else "R2_S_three_prime"
            else:
                metric = "R2_S_five_prime" if read.is_read1 else "R1_S_three_prime"

            if category == "S":
                statistics[metric] = count
            else:
                statistics[metric] = 0

        # Tally up all standard categories from BAM
        if category in statistics:
            statistics[category] += count
        else:
            statistics[category] = count

        # Detect 3 prime soft-clipping
        if i == len(cigar_tuples):
            if strand == cr_constants.REVERSE_STRAND:
                metric = "R2_S_five_prime" if read.is_read2 else "R1_S_three_prime"
            else:
                metric = "R1_S_five_prime" if read.is_read2 else "R2_S_three_prime"

            if category == "S":
                statistics[metric] = count
            else:
                statistics[metric] = 0

    return statistics


def get_full_alignment_base_quality_scores(read: AlignedSegment) -> Tuple[int, np.ndarray]:
    """
    Returns base quality scores for the full read alignment.

    Inserts zeroes for deletions and removing inserted and soft-clipped bases.
    Therefore, only returns quality for truly aligned sequenced bases.

    Args:
        read (pysam.AlignedSegment): read to get quality scores for

    Returns:
        np.array: numpy array of quality scores
    """

    # pylint: disable=no-member
    quality_scores = np.fromstring(read.qual, dtype=np.byte) - tk_constants.ILLUMINA_QUAL_OFFSET

    start_pos = 0

    for operation, length in read.cigar:
        operation = _cigar_numeric_to_category_map[operation]

        if operation == "D":
            quality_scores = np.insert(quality_scores, start_pos, [0] * length)
        elif operation == "I" or operation == "S":
            quality_scores = np.delete(quality_scores, np.s_[start_pos : start_pos + length])

        if not operation == "I" and not operation == "S":
            start_pos += length

    return start_pos, quality_scores


def get_unmapped_read_count_from_indexed_bam(bam_file_name: AnyStr) -> int:
    """
    Get number of unmapped reads from an indexed BAM file.

    Args:
        bam_file_name (str): Name of indexed BAM file.

    Returns:
        int: number of unmapped reads in the BAM

    Note:
        BAM must be indexed for lookup using samtools.
    """

    index_output = tk_subproc.check_output(["samtools", "idxstats", bam_file_name])
    return int(index_output.strip().split(b"\n")[-1].split()[-1])


def kwargs_to_command_line_options(
    reserved_arguments: Union[List[str], Set[str]] = None,
    sep: str = " ",
    long_prefix: str = "--",
    short_prefix: str = "-",
    replace_chars: Optional[Dict[str, str]] = None,
    **kwargs,
) -> List[str]:
    """
    Convert arguments provided by user to a string usable in command line arguments.

    Args:
        reserved_arguments (set or list of str): set of arguments that this function prohibited for user use
        sep (str): separator between option/value pairs (`'='` for `--jobmode=sge`).
                   WARNING: switch args (`--abc`), which take no value, break if sep is not `' '`
        long_prefix (str): prefix for options with more than one character (`"--"` for `--quiet`, for example)
        short_prefix (str): prefix for options with one character (`"-"` for `-q`, for example)
        replace_chars (dict): map of characters to replace in specified variable names
                            (if `--align-reads` is command-line option, specify `align_reads` with `replace_chars` -> `{'_':'-'}`
        **kwargs (dict): `**kwargs` arguments/values to format string for

    Returns:
        list of str: Formatted arguments for use on a command line option.

    Raises:
        ValueError: if user requested argument conflicts with one of the
                    specified reserved arguments.
    """

    arguments = []
    reserved_arguments = set(reserved_arguments) if reserved_arguments else []
    replace_chars = replace_chars.items() if isinstance(replace_chars, dict) else []

    for key, value in kwargs.items():
        normalized_key = key.strip("-")

        # Replace characters for formatting
        for char, new_char in replace_chars:
            normalized_key = normalized_key.replace(char, new_char)

        # Validate user inputs to make sure no blatant conflicts
        if normalized_key in reserved_arguments:
            raise ValueError(
                "Specified option conflicts with reserved argument: %s. \
                             Reserved arguments are: %s"
                % (normalized_key, ",".join(reserved_arguments))
            )

        # Correctly prefix arguments
        if len(key) > 1:
            prefix = long_prefix
        else:
            prefix = short_prefix

        argument = "%s%s" % (prefix, normalized_key)
        option_value = value or ""
        arguments.append("%s%s%s" % (argument, sep, option_value))

    return arguments


def load_barcode_csv(barcode_csv: Union[AnyStr, os.PathLike]) -> Dict[bytes, List[bytes]]:
    """ Load a csv file of (genome,barcode) """

    bcs_per_genome = collections.defaultdict(
        list
    )  # type: collections.defaultdict[bytes, List[bytes]]
    with open(barcode_csv, "r") as f:
        reader = csv.reader(f)
        for row in reader:
            if len(row) != 2:
                raise ValueError("Bad barcode file: %s" % barcode_csv)
            (genome, barcode) = row
            bcs_per_genome[genome.encode()].append(barcode.encode())
    return bcs_per_genome


def get_cell_associated_barcode_set(
    barcode_csv_filename: Union[AnyStr, os.PathLike], genome: Optional[bytes] = None
) -> Set[bytes]:
    """Get set of cell-associated barcode strings.

    Args:
      genome (bytes): Only get cell-assoc barcodes for this genome. If None, disregard genome.

    Returns:
      set of bytes: Cell-associated barcode strings (seq and gem-group).
    """
    cell_bcs_per_genome = load_barcode_csv(barcode_csv_filename)  # type: Dict[bytes, List[bytes]]
    cell_bcs = set()  # type: Set[bytes]
    for g, bcs in cell_bcs_per_genome.items():
        if genome is None or g == ensure_binary(genome):
            cell_bcs |= set(bcs)
    return cell_bcs


def string_is_ascii(input_string: AnyStr) -> bool:
    """Input strings are often stored as ascii in numpy arrays, and we need
    to check that this conversion works."""
    try:
        np.array(ensure_text(input_string, "utf-8"), dtype=bytes)
        return True
    except:
        return False


def chunk_reference(
    input_bam: AlignmentFile, nchunks: int, include_unmapped: bool
) -> List[List[Tuple[str, int, int]]]:
    """Chunk up a reference into nchunks roughly equally sized chunks.

    Args:
        input_bam (pysam.AlignmentFile): Source for reference contig names and lengths.
        nchunks (int): The number of chunks to create.
        include_unmapped (bool): If `True` then one of the chunks consists
                                 of all the unmapped reads as specified by
                                 `[("*", None, None)]`.

    Returns:
        list: loci, where each loci is a list of contiguous regions
              `(contig, start, end)`.
    """
    # one chunk is for unmapped reads
    nchunks -= int(include_unmapped)
    assert nchunks >= 1

    chroms = input_bam.references
    chrom_lengths = input_bam.lengths
    genome_size = sum(chrom_lengths)
    chunk_size = int(np.ceil(float(genome_size) / nchunks))
    process = [(chrom, 0, end) for chrom, end in zip(chroms, chrom_lengths)]

    chunks = []  # type: List[List[Tuple[str, int, int]]]
    if include_unmapped:
        chunks = [[("*", 0, 0)]]
    chunks.append([])

    current_chunk_size = 0
    while len(process):
        piece = process.pop()
        piece_size = piece[2] - piece[1]
        if current_chunk_size + piece_size < chunk_size:
            chunks[-1].append(piece)
            current_chunk_size += piece_size
        else:
            piece_left_over = current_chunk_size + piece_size - chunk_size
            new_piece = (piece[0], piece[1], piece[2] - piece_left_over)
            chunks[-1].append(new_piece)
            chunks.append([])
            current_chunk_size = 0
            if piece_left_over > 0:
                process.append((piece[0], piece[2] - piece_left_over, piece[2]))
    return [c for c in chunks if len(c)]



#!/usr/bin/env python
#
# Copyright (c) 2017 10X Genomics, Inc. All rights reserved.
#

"""Types for loading, saving, and using feature reference data."""

from __future__ import absolute_import, annotations

import os
from collections import OrderedDict

from typing import (
    Collection,
    Dict,
    Generator,
    ItemsView,
    KeysView,
    List,
    Iterable,
    Optional,
    Tuple,
    NamedTuple,
    Union,
    Set,
)

import h5py
import pandas as pd
from six import ensure_binary, ensure_str

import bmkcc.cellranger.hdf5 as cr_h5
import bmkcc.cellranger.h5_constants as h5_constants

FeatureDef = NamedTuple(  # pylint: disable=invalid-name
    "FeatureDef",
    [
        ("index", int),
        ("id", bytes),
        ("name", Optional[str]),
        ("feature_type", str),
        ("tags", Dict[str, str]),
    ],
)

# Required HDF5 datasets
REQUIRED_DATASETS = ["id", "name", "feature_type"]

# These feature tag keys are reserved for internal use.
GENOME_FEATURE_TAG = "genome"
DEFAULT_FEATURE_TAGS = [GENOME_FEATURE_TAG]
RESERVED_TAGS = DEFAULT_FEATURE_TAGS


class FeatureDefException(Exception):
    """Exception type for trying to create a `FeatureReference` with non-distinct IDs."""

    def __init__(self, msg):
        super(FeatureDefException, self).__init__(msg)
        self.msg = msg

    def __str__(self):
        return self.msg


class TargetFeatureSets(object):
    """A set of target features.

    Each target set should be a list of integers with a dictionary key representing the name.
    This name will also exist in the library_info.
    """

    def __init__(self, target_features):
        # type: (Union[None, TargetFeatureSets, Dict[str, Iterable[int]]]) -> None
        self.target_feature_sets = dict()  # type: Dict[str, Set[int]]
        if target_features is not None:
            for name, _set in target_features.items():
                assert all(isinstance(x, int) for x in _set)
                _set = set(_set)
                self.target_feature_sets[name] = _set

    def __eq__(self, other):
        if not isinstance(other, TargetFeatureSets):
            return False
        return self.target_feature_sets == other.target_feature_sets

    def __hash__(self):
        return hash(self.target_feature_sets)

    @property
    def names(self) -> KeysView[str]:
        return self.target_feature_sets.keys()

    def items(self) -> ItemsView[str, Set[int]]:
        return self.target_feature_sets.items()

    def iteritems(self) -> ItemsView[str, Set[int]]:
        """Returns a view of the items.

        .. deprecated::4.0
            Use `items` instead.  Python 3 doesn't do `iteritems`.
        """
        return self.target_feature_sets.items()

    def get(self, key: str) -> Optional[Set[int]]:
        return self.target_feature_sets.get(key)

    def update(self, other: Dict[str, Set[int]]) -> None:
        self.target_feature_sets.update(other)

    def get_target_feature_indices(self) -> List[int]:
        return sorted(set.union(*self.target_feature_sets.values()))

    def subset_to_reduced_features(
        self, old_to_new_index_translation: Dict[int, int]
    ) -> TargetFeatureSets:
        """Recreates a list of target features by subsetting it.

        If the Feature Reference holding this object is ever subset, or the index values are changed
        we need to update the target_feature_sets to hold these new values instead.  This function
        recreates a list of target features by subsetting it to only those present in a translation
        table that maps old index positions to new ones.

        Args:
            old_to_new_index_translation: a dictionary that tells what old index position is being converted
            to a new index position

        Returns:
            A new TargetFeatureSets
        """
        assert isinstance(old_to_new_index_translation, dict)
        new_target_features = {}
        for name, indices in self.target_feature_sets.items():
            new_vals = [
                old_to_new_index_translation[old_index]
                for old_index in indices
                if old_index in old_to_new_index_translation
            ]
            new_target_features[name] = new_vals
        return TargetFeatureSets(new_target_features)


class FeatureReference(object):
    """Store a list of features (genes, antibodies, etc)."""

    def __init__(self, feature_defs, all_tag_keys, target_features=None):
        # type: (List[FeatureDef], List[str], Union[None, TargetFeatureSets, Dict[str, Iterable[int]]]) -> None
        """Create a FeatureReference.

        Args:
            feature_defs (list of FeatureDef): All feature definitions.
            all_tag_keys (list of str): All optional tag keys.
            target_features (dictionary of list of int): Optional target set(s). Each target set
            should be a list of integers with a dictionary key representing the name. This name
            will also exist in the library_info.
        """
        self.feature_defs = feature_defs
        self.all_tag_keys = all_tag_keys

        if target_features is not None:
            self.target_features = TargetFeatureSets(target_features)
        else:
            self.target_features = None

        # Assert uniqueness of feature IDs
        id_map = {}
        for fdef in self.feature_defs:
            if fdef.id in id_map:
                this_fd_str = "ID: %s; name: %s; type: %s" % (fdef.id, fdef.name, fdef.feature_type)
                seen_fd_str = "ID: %s; name: %s; type: %s" % (
                    id_map[fdef.id].id,
                    id_map[fdef.id].name,
                    id_map[fdef.id].feature_type,
                )
                raise FeatureDefException(
                    "Found two feature definitions with the same ID: "
                    "(%s) and (%s). All feature IDs must be distinct." % (this_fd_str, seen_fd_str)
                )
            id_map[fdef.id] = fdef

        self.id_map = id_map  # type: Dict[bytes, FeatureDef]

    def get_count_of_feature_type(self, feature_type: str) -> int:
        """ Count how many features in the matrix are of a given type. (e.g. "Gene Expression")"""
        total = 0
        for feature in self.feature_defs:
            if feature.feature_type == feature_type:
                total += 1
        return total

    def __eq__(self, other):
        return (
            self.feature_defs == other.feature_defs
            and self.all_tag_keys == other.all_tag_keys
            and (self.target_features is None) == (other.target_features is None)
            and self.target_features == other.target_features
        )

    def __hash__(self):
        if self.target_features is None:
            return hash((self.feature_defs, self.all_tag_keys))
        return hash((self.feature_defs, self.all_tag_keys, self.target_features))

    def __ne__(self, other):
        return not self == other

    def equals_ignore_target_set(self, other: FeatureReference) -> bool:
        """Checks if two feature references are equal, ignoring missing targets.

        Only takes into account equality of the
        target set if it is present in both feature references. Useful for targeted feature reference
        compatibility checks in preflights.
        """
        ignore_target_set = (not self.has_target_features()) or (not other.has_target_features())
        compatible_target_sets = ignore_target_set or (
            self.get_target_feature_indices() == other.get_target_feature_indices()
        )
        return (
            self.feature_defs == other.feature_defs
            and self.all_tag_keys == other.all_tag_keys
            and compatible_target_sets
        )

    @staticmethod
    def addtags(
        feature_ref: FeatureReference,
        new_tags: List[str],
        new_labels: Optional[Collection[Collection[str]]] = None,
    ) -> FeatureReference:
        """Add new tags and corresponding labels to existing feature_ref.

        If new labels are None, empty strings are supplied by default

        Args:
            feature_ref: a FeatureReference instance
            new_tags: a list of new tags
            new_labels: per feature list of label values corresponding to the new tags
        """
        assert len(new_tags) > 0
        for tag in new_tags:
            if tag in feature_ref.all_tag_keys:
                raise ValueError("tag {} is already present in feature_ref")

        if new_labels is not None:
            if len(feature_ref.feature_defs) != len(new_labels):
                raise ValueError(
                    "number of labels does not match number of features " "in feature_ref"
                )
            for labels in new_labels:
                assert len(labels) == len(new_tags)
            use_labels = new_labels
        else:
            # initialize to empty
            use_labels = [[""] * len(new_tags) for _ in range(len(feature_ref.feature_defs))]
        assert len(feature_ref.feature_defs) == len(use_labels)

        augmented_features = []
        for f_def, newvals in zip(feature_ref.feature_defs, use_labels):
            tags = dict(zip(new_tags, newvals))
            tags.update(f_def.tags)
            augmented_features.append(
                FeatureDef(
                    index=f_def.index,
                    id=f_def.id,
                    name=f_def.name,
                    feature_type=f_def.feature_type,
                    tags=tags,
                )
            )

        return FeatureReference(
            feature_defs=augmented_features,
            all_tag_keys=feature_ref.all_tag_keys + new_tags,
            target_features=feature_ref.target_features,
        )

    @staticmethod
    def join(feature_ref1: FeatureReference, feature_ref2: FeatureReference) -> FeatureReference:
        """Concatenate two feature references, requires unique ids and identical tags"""
        assert feature_ref1.all_tag_keys == feature_ref2.all_tag_keys
        feature_defs1 = feature_ref1.feature_defs
        feature_defs2 = feature_ref2.feature_defs

        if feature_ref1.target_features is None:
            combined_target_features = feature_ref2.target_features
        elif feature_ref2.target_features is None:
            combined_target_features = feature_ref1.target_features
        else:
            combined_target_features = feature_ref1.target_features
            # if feature_ref2 has the same keys, they will be over-written
            combined_target_features.update(feature_ref2.target_features)
        return FeatureReference(
            feature_defs=feature_defs1 + feature_defs2,
            all_tag_keys=feature_ref1.all_tag_keys,
            target_features=combined_target_features,
        )

    @classmethod
    def empty(cls) -> FeatureReference:
        return cls(feature_defs=[], all_tag_keys=[], target_features=None)

    def get_num_features(self) -> int:
        return len(self.feature_defs)

    def get_feature_ids_by_type(self, feature_type: str) -> List[bytes]:
        """ Return a list of feature ids of a particular feature type (e.g. "Gene Expression")"""
        return [f.id for f in self.feature_defs if f.feature_type == feature_type]

    def get_indices_for_type(self, feature_type: str) -> List[int]:
        return [
            feature.index for feature in self.feature_defs if feature.feature_type == feature_type
        ]

    def get_genomes(self, feature_type: Optional[str] = None) -> List[str]:
        """Get sorted list of genomes. Empty string is for reverse compatibility.

        A specific feature type can optionally be specified.
        """
        genomes = set(
            f.tags.get(GENOME_FEATURE_TAG, "")
            for f in self.feature_defs
            if (feature_type is None or f.feature_type == feature_type)
        )
        return sorted(genomes)

    def has_target_features(self) -> bool:
        return self.target_features is not None

    def get_target_feature_indices(self) -> Optional[List[int]]:
        """Gets the indices of on-target features within the FeatureRerence.

        Returns None if there is no target set.
        """
        if not self.has_target_features():
            return None
        else:
            return self.target_features.get_target_feature_indices()

    def get_target_feature_ids(self) -> Optional[List[bytes]]:
        """Gets the feature ids of on-target features.

        Returns None if there is no target set.
        """
        if not self.has_target_features():
            return None
        else:
            return [self.feature_defs[i].id for i in self.get_target_feature_indices()]

    def select_features_by_type(self, feature_type: str) -> FeatureReference:
        indices = [i for i, fd in enumerate(self.feature_defs) if fd.feature_type == feature_type]
        return self.select_features(indices)

    def get_feature_names(self) -> List[str]:
        """Get a list of feature names.

        :return: [x.name for x in self.feature_defs]
        """
        return [x.name for x in self.feature_defs]

    def select_features(self, indices: Iterable[int]) -> FeatureReference:
        """Create a new FeatureReference that only contains features with the given indices.

        Any target sets present are updated, but the target_features attribute in the output
        is set to None if all target features were removed.
        """
        old_defs = [self.feature_defs[i] for i in indices]
        new_defs = []  # type: List[FeatureDef]
        translation_map = {}
        for i, f_def in enumerate(old_defs):
            translation_map[f_def.index] = i
            new_defs.append(
                FeatureDef(
                    index=i,
                    id=f_def.id,
                    name=f_def.name,
                    feature_type=f_def.feature_type,
                    tags=f_def.tags,
                )
            )

        if self.has_target_features():
            new_target_features = self.target_features.subset_to_reduced_features(translation_map)
            # If we have removed all targeted features, we toss the empty target sets
            if len(new_target_features.get_target_feature_indices()) == 0:
                new_target_features = None
        else:
            new_target_features = None
        return FeatureReference(
            feature_defs=new_defs,
            all_tag_keys=self.all_tag_keys,
            target_features=new_target_features,
        )

    def to_hdf5(self, group: h5py.Group) -> None:
        """Write to an HDF5 group."""

        # Write required datasets
        for col in REQUIRED_DATASETS:
            data = [getattr(f, col) for f in self.feature_defs]
            cr_h5.create_hdf5_string_dataset(group, col, data, compression=True)

        # Write tag datasets
        for col in self.all_tag_keys:
            # Serialize missing data as empty unicode string
            data = [f.tags.get(col, "") for f in self.feature_defs]
            cr_h5.create_hdf5_string_dataset(group, col, data, compression=True)

        # Write target_features as a new sub-group
        if self.target_features is not None:
            target_set_group = group.create_group(h5_constants.H5_TARGET_SET_ATTR)
            for key, val in self.target_features.items():
                cr_h5.create_hdf5_string_dataset(
                    target_set_group, key, [str(x) for x in sorted(val)], compression=True
                )

        # Record names of all tag columns
        cr_h5.create_hdf5_string_dataset(group, "_all_tag_keys", self.all_tag_keys)

    @classmethod
    def from_hdf5(cls, group: h5py.Dataset) -> FeatureReference:
        """Load from an HDF5 group.

        Args:
            group (h5py.Dataset): Group to load from.

        Returns:
            feature_ref (FeatureReference): New object.
        """
        # FIXME: ordering may not be guaranteed in python3
        def _load_str(node: h5py.Dataset) -> List[str]:
            if node.shape is None:
                return []
            if node.dtype.char == cr_h5.STR_DTYPE_CHAR:
                memoize = node.name in ("/features/feature_type", "/features/genome")
                return cr_h5.read_hdf5_string_dataset(node, memoize)
            else:
                return node[:]

        def _format_path(path) -> str:
            """ Strip off leading slash and convert to str (could be unicode). """
            return ensure_str(path[1:])

        def _h5py_dataset_iterator(
            group: h5py.Group, prefix: str = ""
        ) -> Generator[Tuple[str, List[str]], None, None]:
            for key in group:
                item = group[key]
                path = "/".join([prefix, key])
                if isinstance(item, h5py.Dataset):
                    yield _format_path(path), _load_str(item)
                elif isinstance(item, h5py.Group):
                    # TODO: in python 3.3 this can be a yield from statement
                    for subpath, subitem in _h5py_dataset_iterator(item, path):
                        yield subpath, subitem

        data = OrderedDict(_h5py_dataset_iterator(group))

        # Load Tag Keys
        all_tag_keys = data["_all_tag_keys"]

        # Load Target Sets, if they exist
        target_features = {
            os.path.basename(key): [int(x) for x in val]
            for key, val in data.items()
            if h5_constants.H5_TARGET_SET_ATTR in ensure_binary(key)
        }
        if len(target_features) == 0:
            target_features = None

        # Build FeatureDefs
        feature_defs = []  # List[FeatureDef]
        num_features = len(data[REQUIRED_DATASETS[0]])

        for i in range(num_features):
            tags = {ensure_str(k): ensure_str(data[k][i]) for k in all_tag_keys if data[k][i]}
            feature_defs.append(
                FeatureDef(
                    id=ensure_binary(data["id"][i]),
                    index=i,
                    name=data["name"][i],
                    feature_type=data["feature_type"][i],
                    tags=tags,
                )
            )

        return cls(feature_defs, all_tag_keys=all_tag_keys, target_features=target_features)


class SummaryFeatureCounts(object):
    """Stores and queries a dataframe that summarizes relevant features per gene."""

    def __init__(self, df: pd.DataFrame, filepath=None):
        """Stores and queries a dataframe that summarizes relevant features per gene.

        There should be one row per gene and any number of columns.

        Args:
            df (pandas dataframe): Dataframe with n_genes rows and columns with features per genes.
        """
        self.df = df
        # for easy querying
        self.feature_dict = df.set_index("feature_id").to_dict(orient="index")
        self.filepath = filepath

    @classmethod
    def from_file(cls, path) -> SummaryFeatureCounts:
        """Loads the features from a csv file."""
        path = ensure_str(path)
        assert os.path.isfile(path)
        return cls(pd.read_csv(ensure_str(path)), path)

    def to_csv(self, path):
        path = ensure_str(path)
        self.df.to_csv(path, index=False)

    def get_df(self) -> pd.DataFrame:
        return self.df

    def get_value_for_feature(self, feature_id, key):
        """Get the given key for the given feature."""
        if feature_id not in self.feature_dict:
            return None
        if key not in self.feature_dict[feature_id]:
            return None
        return self.feature_dict[feature_id][key]



#!/usr/bin/env python
#
# Copyright (c) 2016 10X Genomics, Inc. All rights reserved.
#
# Manage data used by the webshim

from __future__ import absolute_import, division

from six.moves import xrange as range

import h5py
import json
import pandas as pd
from six import ensure_binary, ensure_str

#import cellranger.analysis.singlegenome as cr_sg_analysis
#import cellranger.analysis.multigenome as cr_mg_analysis
from bmkcc.cellranger.feature_ref import SummaryFeatureCounts
import bmkcc.cellranger.webshim.constants.gex as gex_constants
import bmkcc.cellranger.utils as cr_utils
#import cellranger.vdj.utils as vdj_utils
'''
from cellranger.websummary.sample_properties import (
    SampleProperties,
    CountSampleProperties,
    ExtendedCountSampleProperties,
)
'''
import collections
import numpy as np

BarcodeRankPlotSegment = collections.namedtuple(
    "BarcodeRankPlotSegment", ["start", "end", "cell_density", "legend"]
)


def get_plot_segment(start_index, end_index, sorted_bc, cell_barcodes, legend=False):
    """
    Helper function to build a plot segment.
    """
    assert end_index > start_index
    num_cells = sum(1 for i in range(start_index, end_index) if sorted_bc[i] in cell_barcodes)
    density = float(num_cells) / float(end_index - start_index)
    return BarcodeRankPlotSegment(
        start=start_index, end=end_index, cell_density=density, legend=legend
    )


def generate_counter_barcode_rank_plot_data(cell_barcodes, barcode_summary, key):
    """
    A helper function to generate the data required to generate the barcode rank
    plot for RNA counter. The barcode rank plot consists of
    multiple plot segments

    If the barcodes are ordered by UMI counts
        - All the barcodes until the first non cell barcodes go into
        the `Cell` segment.
        - All the barcodes beyond the last cell barcode goes into the
        `Background` segment.
        - The remaining barcodes are further divided into multiple plot
        segments
    Input:
        - cell_barcodes: A set of bytes with the cell barcodes
        - barcode_summary: The barcode summary from the h5.
        - key: Specifies the entry to look up in barcode summary
            for getting the UMI counts
    Output:
        - sorted_counts: UMI counts sorted in descending order
        - plot_segments: List of BarcodeRankPlotSegment
    """
    counts_per_bc = barcode_summary[key][:]
    srt_order = compute_sort_order(counts_per_bc, barcode_summary["bc_sequence"][:], cell_barcodes)
    sorted_bc = barcode_summary["bc_sequence"][:][srt_order]
    sorted_counts = counts_per_bc[srt_order]
    del srt_order

    plot_segments = compute_plot_segments(sorted_bc, sorted_counts, cell_barcodes)

    return sorted_counts, plot_segments


class SampleData:
    def __init__(self, sample_properties, sample_data_paths, plot_preprocess_func):
        if sample_properties:  # Can be None
            assert isinstance(sample_properties, SampleProperties)
        self.summary = load_metrics_summary(sample_data_paths.summary_path)

        # Not guaranteed to exist for all pipelines
        self.num_cells = self.summary.get("filtered_bcs_transcriptome_union")

        self.analyses, self.original_cluster_sizes = load_analyses(
            sample_data_paths.analysis_path, plot_preprocess_func, sample_properties
        )

        self.barcode_summary = load_barcode_summary(sample_data_paths.barcode_summary_path)

        self.has_target_set = (
            isinstance(sample_properties, ExtendedCountSampleProperties)
            and sample_properties.target_set is not None
        )

        self.vdj_clonotype_summary = (
            pd.read_csv(ensure_str(sample_data_paths.vdj_clonotype_summary_path))
            if sample_data_paths.vdj_clonotype_summary_path
            else None
        )

        self.vdj_barcode_support = (
            pd.read_csv(
                ensure_str(sample_data_paths.vdj_barcode_support_path),
                converters={"barcode": ensure_binary},
            )
            if sample_data_paths.vdj_barcode_support_path
            else None
        )

        self.cell_barcodes = (
            cr_utils.get_cell_associated_barcode_set(sample_data_paths.filtered_barcodes_path)
            if sample_data_paths.filtered_barcodes_path
            else None
        )

        if sample_data_paths.vdj_cell_barcodes_path:
            assert self.cell_barcodes is None
            self.cell_barcodes = {
                ensure_binary(bc)
                for bc in vdj_utils.load_cell_barcodes_json(
                    sample_data_paths.vdj_cell_barcodes_path
                )
            }

        self.feature_metrics = load_feature_counts(sample_data_paths.feature_metrics_path)

    def get_analysis(self, analysis_type):
        if self.analyses is None:
            return None
        r = [a for a in self.analyses if isinstance(a, analysis_type)]
        assert len(r) == 0 or len(r) == 1
        if len(r) == 0:
            return None
        else:
            return r[0]

    def counter_barcode_rank_plot_data(self, key):
        """
        Generate the data required to generate the barcode rank
        plot for RNA counter. The barcode rank plot consists of
        multiple plot segments

        If the barcodes are ordered by UMI counts
            - All the barcodes until the first non cell barcodes go into
              the `Cell` segment.
            - All the barcodes beyond the last cell barcode goes into the
              `Background` segment.
            - The remaining barcodes are further divided into multiple plot
              segments
        Input:
            - key: Specifies the entry to look up in barcode summary
                   for getting the UMI counts
        Output:
            - sorted_counts: UMI counts sorted in descending order
            - plot_segments: List of BarcodeRankPlotSegment
        """

        return generate_counter_barcode_rank_plot_data(
            self.cell_barcodes, self.barcode_summary, key
        )

    def vdj_barcode_rank_plot_data(self):
        """
        Generate the data required to generate the barcode rank plot for VDJ.
        """
        counts_per_bc = self.vdj_barcode_support["count"].to_numpy()
        srt_order = compute_sort_order(
            counts_per_bc, self.vdj_barcode_support["barcode"], self.cell_barcodes
        )
        sorted_bc = self.vdj_barcode_support["barcode"].to_numpy()[srt_order]
        sorted_counts = counts_per_bc[srt_order]
        del srt_order

        plot_segments = compute_plot_segments(sorted_bc, sorted_counts, self.cell_barcodes)
        return sorted_counts, plot_segments

    def is_targeted(self):
        return self.has_target_set


def compute_sort_order(counts_per_bc, bc_sequences, cell_barcodes):
    assert len(counts_per_bc) == len(bc_sequences)
    is_cell = np.full(len(counts_per_bc), False)
    for i, bc in enumerate(bc_sequences):
        is_cell[i] = bc in cell_barcodes
    srt_order = sorted(
        range(len(counts_per_bc)), key=lambda x: (counts_per_bc[x], is_cell[x]), reverse=True
    )
    return srt_order


def compute_plot_segments(sorted_bc, sorted_counts, cell_barcodes):
    # find the first barcode which is not a cell
    first_non_cell = len(sorted_bc)
    for i, bc in enumerate(sorted_bc):
        if bc not in cell_barcodes:
            first_non_cell = i
            break

    # find the last barcode which is a cell
    last_cell = 0
    for i in reversed(range(len(sorted_bc))):
        if sorted_bc[i] in cell_barcodes:
            last_cell = i
            break

    ranges = [0, first_non_cell, last_cell + 1, len(sorted_bc)]

    plot_segments = []
    plot_segments.append(
        BarcodeRankPlotSegment(start=0, end=ranges[1], cell_density=1.0, legend=True)
    )
    plot_segments.append(
        BarcodeRankPlotSegment(start=ranges[2], end=ranges[3], cell_density=0.0, legend=True)
    )

    # Subdivide the mixed section
    mixed_segments = segment_log_plot_by_length(sorted_counts, ranges[1], ranges[2])
    for i in range(len(mixed_segments) - 1):
        plot_segments.append(
            get_plot_segment(
                mixed_segments[i], mixed_segments[i + 1], sorted_bc, cell_barcodes, legend=False
            )
        )

    return plot_segments


def segment_log_plot_by_length(y_data, x_start, x_end):
    """
    Given the extends of the mixed region [x_start, x_end), compute
    the x-indices that would divide the plot into segments of a
    prescribed length (in pixel coordinates) with a minimum number
    of barcodes in each segment
    """

    if x_end <= x_start:
        return []

    SEGMENT_NORMALIZED_MAX_LEN = 0.02
    MIN_X_SPAN = 20

    log_max_x = np.log(len(y_data))
    log_max_y = np.log(max(y_data))

    this_segment_len = 0.0
    segment_idx = [x_start]

    for i in range(x_start, x_end):
        if i == 0:
            continue
        last_i = max(x_start, i - 1)
        dx = (np.log(i) - np.log(last_i)) / log_max_x
        dy = (np.log(y_data[i]) - np.log(y_data[last_i])) / log_max_y
        this_segment_len += np.linalg.norm([dx, dy])
        if this_segment_len >= SEGMENT_NORMALIZED_MAX_LEN and i > (segment_idx[-1] + MIN_X_SPAN):
            segment_idx.append(i + 1)
            this_segment_len = 0.0

    if segment_idx[-1] != x_end:
        segment_idx.append(x_end)

    return segment_idx


def load_metrics_summary(summary_path):
    if summary_path is None:
        return None
    with open(summary_path) as f:
        return json.load(f)


def load_feature_counts(feature_metrics_path):
    if feature_metrics_path is None:
        return None
    return SummaryFeatureCounts.from_file(feature_metrics_path)


def load_analyses(base_dir, plot_preprocess_func, sample_properties):
    """ Returns (analysis_object, original_cluster_sizes) """
    if base_dir is None:
        return None, None
    if isinstance(sample_properties, CountSampleProperties):
        if len(sample_properties.genomes) == 1:
            analyses = [cr_sg_analysis.SingleGenomeAnalysis.load_default_format(base_dir, "pca")]
        else:
            analyses = [
                cr_sg_analysis.SingleGenomeAnalysis.load_default_format(base_dir, "pca"),
                cr_mg_analysis.MultiGenomeAnalysis.load_default_format(base_dir),
            ]

    if analyses[0] is None:
        return None, None

    # Subsample barcodes for plotting purposes
    original_cluster_sizes = None
    analysis = analyses[0]
    if analysis is not None and isinstance(analysis, cr_sg_analysis.SingleGenomeAnalysis):
        original_cluster_sizes = analysis.get_cluster_sizes()
        if analysis.matrix.bcs_dim > gex_constants.MAX_WEBSHIM_BCS_DIM:
            analysis.subsample_bcs(gex_constants.MAX_WEBSHIM_BCS_DIM)

    return plot_preprocess_func(analyses), original_cluster_sizes


def load_barcode_summary(h5_path):
    if h5_path is None:
        return None
    return h5py.File(h5_path, "r")



#!/usr/bin/env python
#
# Copyright (c) 2016 10X Genomics, Inc. All rights reserved.
#
from __future__ import absolute_import

import csv
import numpy as np
import os
import six
from bmkcc.cellranger.wrapped_tables import tables
import bmkcc.cellranger.h5_constants as h5_constants

# Version for HDF5 format
VERSION_KEY = "version"
VERSION = 2


def save_h5(f, group, key, namedtuple):
    """ Save a namedtuple to an h5 file under a group and subgroup """
    if VERSION_KEY in f.root:
        version = int(getattr(f.root, VERSION_KEY))
        if version != VERSION:
            raise ValueError(
                "Attempted to write analysis HDF5 version %d data to a version %d file"
                % (VERSION, version)
            )
    else:
        ds = f.create_array(f.root, VERSION_KEY, np.int64(VERSION))

    subgroup = f.create_group(group, "_" + key)
    for field in namedtuple._fields:
        arr = getattr(namedtuple, field)

        # XML encode strings so we can store them as HDF5 ASCII
        if isinstance(arr, six.text_type):
            arr = np.string_(arr.encode("ascii", "xmlcharrefreplace"))
        elif isinstance(arr, bytes):
            arr = np.string_(arr)

        if not hasattr(arr, "dtype"):
            raise ValueError("%s/%s must be a numpy array or scalar" % (group, key))

        atom = tables.Atom.from_dtype(arr.dtype)
        if len(arr.shape) > 0:
            if arr.size > 0:
                ds = f.create_carray(subgroup, field, atom, arr.shape)
            else:
                ds = f.create_earray(subgroup, field, atom, arr.shape)
            ds[:] = arr
        else:
            ds = f.create_array(subgroup, field, arr)


def _string_or_num(value):
    if isinstance(value, str):
        return value
    elif isinstance(value, (bytes, six.text_type)):
        return six.ensure_str(value)
    return value


def _string_or_num_list(value):
    # type: (...) -> list
    """Convert a string or collection into a list of str or numbers.

    `bytes` are decoded to `str`s.

    `str` or numbers are wrapped in lists.

    For iterables (other than strings), a list is returned where any `bytes`
    elements have been decoded to `str`.

    Args:
        value (str, bytes, int, float, iterable): The value to convert.

    Returns:
        list: A list containing `str` or numeric values.
    """
    if not isinstance(value, (bytes, six.text_type)) and hasattr(value, "__iter__"):
        return [_string_or_num(v) for v in value]
    return [_string_or_num(value)]


def save_matrix_csv(filename, arr, header, prefixes):
    with open(filename, "w") as f:
        writer = csv.writer(f, lineterminator="\n")
        writer.writerow(header)
        # Iterate over the given arr, default iteration is by-row
        for v, prefix in zip(arr, prefixes):
            row = _string_or_num_list(prefix)
            row.extend(_string_or_num_list(v))
            writer.writerow(row)


def load_h5_namedtuple(group, namedtuple):
    """ Load a single namedtuple from an h5 group """
    args = []
    for field in namedtuple._fields:
        try:
            field_value = getattr(group, field).read()
            if field_value.shape == ():
                field_value = field_value.item()
        except tables.NoSuchNodeError:
            try:
                field_value = getattr(group._v_attrs, field)
            except AttributeError:
                field_value = None
        args.append(field_value)
    return namedtuple(*args)


def load_h5_iter(group, namedtuple):
    for subgroup in group:
        yield subgroup._v_name[1:], load_h5_namedtuple(subgroup, namedtuple)


def h5_path(base_path):
    return os.path.join(base_path, "analysis.h5")


def _combine_h5_group(fins, fout, group):
    group_out = fout.create_group(fout.root, group)
    for fin in fins:
        # Skip non-existent input groups
        if group not in fin.root:
            continue

        group_in = fin.root._v_groups[group]

        # NOTE - this throws an exception if a child group already exists
        fin.copy_children(group_in, group_out, recursive=True)


def combine_h5_files(in_files, out_file, groups):
    fins = [tables.open_file(filename, "r") for filename in in_files]
    with tables.open_file(out_file, "w") as fout:
        for group in groups:
            _combine_h5_group(fins, fout, group)
    for fin in fins:
        fin.close()


def open_h5_for_writing(filename):
    filters = tables.Filters(complevel=h5_constants.H5_COMPRESSION_LEVEL)
    return tables.open_file(filename, "w", filters=filters)


def save_dimension_reduction_h5(data_map, fname, group_name):
    filters = tables.Filters(complevel=h5_constants.H5_COMPRESSION_LEVEL)
    with tables.open_file(fname, "w", filters=filters) as f:
        group = f.create_group(f.root, group_name)
        for n_components, reduction in six.iteritems(data_map):
            save_h5(f, group, str(n_components), reduction)


def load_dimension_reduction_from_h5(filename, group, ntuple):
    with tables.open_file(filename, "r") as f:
        group = f.root._v_groups[group]
        # Just take the first object, assuming we never have multiple
        for _, pca in load_h5_iter(group, ntuple):
            return pca



import argparse
from .cells import save_feature, callcells, barcode_umi_plot, writeQC

def parseArgs():
    parser = argparse.ArgumentParser('')
    parser.add_argument('--matrix', '-m', required=True, help='raw feature barcode matrix')
    parser.add_argument('--expect-cells', '-e', default=3000, required=False,
                        type=int, help='number of expected cells')
    parser.add_argument('--basedir', '-d', default='.', help='output direction')
    return parser.parse_args()


def main():
    args = parseArgs()
    expect_cells = args.expect_cells
    basedir = args.basedir
    matrixDir = args.matrix
    matrix, filtered_matrix = callcells(matrixDir, basedir, expect_cells)
    cell_barcodes = filtered_matrix.bcs[:]
    counts_per_bc = matrix.get_counts_per_bc()
    raw_barcodes = matrix.bcs[:]
    barcode_umi_plot(counts_per_bc, raw_barcodes, cell_barcodes, basedir)
    writeQC(filtered_matrix, basedir)
    print('completed!')

if __name__=='__main__':
    main()
